# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from dataclasses import dataclass, field
from typing import List, Optional, Tuple, Callable
import torch
import torch.nn as nn
import torch.nn.functional as F
from transformers.activations import ACT2FN
from transformers.modeling_rope_utils import ROPE_INIT_FUNCTIONS
from transformers import AutoConfig
from cosmos_rl.utils.util import (
    resolve_model_path,
    IdentityLayer,
    clear_weight_name,
    sync_model_vocab,
    retry,
)
from safetensors import safe_open
from cosmos_rl.policy.model.qwen2_5_vl.weight_converter import (
    convert_weight_from_hf,
)
from cosmos_rl.dispatcher.data.packer.qwen2_5_vlm_data_packer import (
    Qwen2_5_VLM_DataPacker,
)
from cosmos_rl.policy.model.qwen2_5_vl.weight_mapper import QwenVL25WeightMapper
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.policy.model.base import ModelRegistry, BaseModel
from functools import cached_property
import cosmos_rl.policy.kernel.modeling_utils as modeling_utils
from cosmos_rl.policy.kernel.norm import RMSNorm
from cosmos_rl.policy.kernel.fused import MLPActMulFunc
from cosmos_rl.utils.sequence_packing import pack_sequences_for_inputs


@dataclass
class Qwen2_5_VL_LM_Args:
    mrope_section: List[int]
    dim: int
    ffn_dim: int
    n_layers: int
    n_heads: int
    n_kv_heads: int
    vocab_size: int
    max_seq_len: int
    biases: List[str] = field(default_factory=lambda: [])
    norm_eps: float = 1e-6
    rope_theta: float = 10000
    norm_type: str = "rmsnorm"
    rope_type: str = "default"
    hidden_act: str = "silu"
    hf_config: AutoConfig = None


@dataclass
class Qwen2_5_VL_Encoder_Args:
    depth: int
    hidden_size: int
    hidden_act: str
    intermediate_size: int
    n_heads: int
    in_channels: int
    patch_size: int
    spatial_merge_size: int
    temporal_patch_size: int
    tokens_per_second: int
    window_size: int
    fullatt_block_indexes: List[int]
    out_hidden_size: int
    norm_type: str = "rmsnorm"
    norm_eps: float = 1e-6
    hf_config: AutoConfig = None


@dataclass
class Qwen2_5_VL_Args:
    lm_args: Qwen2_5_VL_LM_Args
    encoder_args: Qwen2_5_VL_Encoder_Args
    hf_config: AutoConfig = None


"""
Vision part
"""


class Qwen2_5_VLMLP(nn.Module):
    def __init__(self, config: Qwen2_5_VL_LM_Args, bias: bool = False):
        super().__init__()
        self.hidden_size = config.hidden_size
        self.intermediate_size = config.intermediate_size
        self.gate_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=bias)
        self.up_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=bias)
        self.down_proj = nn.Linear(self.intermediate_size, self.hidden_size, bias=bias)
        self.act_mul_func = MLPActMulFunc(ACT2FN[config.hidden_act])

    def forward(self, hidden_state):
        return self.down_proj(
            self.act_mul_func(self.gate_proj(hidden_state), self.up_proj(hidden_state))
        )


class Qwen2_5_VisionPatchEmbed(nn.Module):
    def __init__(
        self,
        patch_size: int = 14,
        temporal_patch_size: int = 2,
        in_channels: int = 3,
        embed_dim: int = 1152,
    ) -> None:
        super().__init__()
        self.patch_size = patch_size
        self.temporal_patch_size = temporal_patch_size
        self.in_channels = in_channels
        self.embed_dim = embed_dim

        kernel_size = [temporal_patch_size, patch_size, patch_size]
        self.proj = nn.Conv3d(
            in_channels,
            embed_dim,
            kernel_size=kernel_size,
            stride=kernel_size,
            bias=False,
        )

    def forward(self, hidden_states: torch.Tensor) -> torch.Tensor:
        target_dtype = self.proj.weight.dtype
        hidden_states = hidden_states.view(
            -1,
            self.in_channels,
            self.temporal_patch_size,
            self.patch_size,
            self.patch_size,
        )
        hidden_states = self.proj(hidden_states.to(dtype=target_dtype)).view(
            -1, self.embed_dim
        )
        return hidden_states


class Qwen2_5_VisionRotaryEmbedding(nn.Module):
    def __init__(self, dim: int, theta: float = 10000.0) -> None:
        super().__init__()
        self.theta = theta
        self.dim = dim
        self.reset_inv_freq()

    def reset_inv_freq(self, device: torch.device = None):
        inv_freq = 1.0 / (
            self.theta ** (torch.arange(0, self.dim, 2, dtype=torch.float) / self.dim)
        )
        inv_freq = inv_freq.to(device) if device is not None else inv_freq
        if not hasattr(self, "inv_freq"):
            self.register_buffer("inv_freq", inv_freq, persistent=False)
        else:
            self.inv_freq.to(torch.float32)
            with torch.no_grad():
                self.inv_freq.data.copy_(inv_freq)

    def forward(self, seqlen: int, device: torch.device = None) -> torch.Tensor:
        if self.inv_freq.dtype != torch.float32:
            self.reset_inv_freq(device=device)
            assert (
                self.inv_freq.dtype == torch.float32
            ), "inv_freq dtype should be float32"
        seq = torch.arange(
            seqlen, device=self.inv_freq.device, dtype=self.inv_freq.dtype
        )
        freqs = torch.outer(seq, self.inv_freq)
        return freqs


class Qwen2_5_VLPatchMerger(nn.Module):
    def __init__(self, config: Qwen2_5_VL_Encoder_Args) -> None:
        super().__init__()
        self.hidden_size = config.hidden_size * (config.spatial_merge_size**2)
        self.ln_q = RMSNorm(
            config.hidden_size,
            config.norm_eps,
            casting_mode=config.hf_config.model_type,
        )
        self.mlp = nn.Sequential(
            nn.Linear(self.hidden_size, self.hidden_size),
            nn.GELU(),  # This is fixed to GELU according to the original implementation
            nn.Linear(self.hidden_size, config.out_hidden_size),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.mlp(self.ln_q(x).view(-1, self.hidden_size))
        return x


def rotate_half(x):
    """Rotates half the hidden dims of the input."""
    x1 = x[..., : x.shape[-1] // 2]
    x2 = x[..., x.shape[-1] // 2 :]
    return torch.cat((-x2, x1), dim=-1)


if os.environ.get("COSMOS_USE_HF_IMPL", "0").lower() in ["1", "true"]:

    def apply_rotary_pos_emb_vision(
        q: torch.Tensor, k: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        orig_q_dtype = q.dtype
        orig_k_dtype = k.dtype
        q, k = q.float(), k.float()
        cos, sin = cos.unsqueeze(-2).float(), sin.unsqueeze(-2).float()
        q_embed = (q * cos) + (rotate_half(q) * sin)
        k_embed = (k * cos) + (rotate_half(k) * sin)
        q_embed = q_embed.to(orig_q_dtype)
        k_embed = k_embed.to(orig_k_dtype)
        return q_embed, k_embed
else:

    def apply_rotary_pos_emb_vision(
        q: torch.Tensor, k: torch.Tensor, cos: torch.Tensor, sin: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Apply rotary position embedding to the query and key tensors.
        """
        cos = cos.chunk(2, dim=-1)[0].contiguous()
        sin = sin.chunk(2, dim=-1)[0].contiguous()
        q_embed = modeling_utils.apply_rotary_emb(
            q.float(), cos.float(), sin.float()
        ).type_as(q)
        k_embed = modeling_utils.apply_rotary_emb(
            k.float(), cos.float(), sin.float()
        ).type_as(k)
        return q_embed, k_embed


class Qwen2_5_VLVisionAttention(nn.Module):
    def __init__(self, dim: int, num_heads: int = 16) -> None:
        super().__init__()
        self.num_heads = num_heads
        self.qkv = nn.Linear(dim, dim * 3, bias=True)
        self.proj = nn.Linear(dim, dim)
        self.attention_dropout = 0.0
        self.attn_func = modeling_utils.flash_attn_varlen_func

    def forward(
        self,
        hidden_states: torch.Tensor,
        cu_seqlens: torch.Tensor,
        position_embeddings: torch.Tensor,
        max_seqlen: int,
    ) -> torch.Tensor:
        seq_length = hidden_states.shape[0]
        q, k, v = (
            self.qkv(hidden_states)
            .reshape(seq_length, 3, self.num_heads, -1)
            .permute(1, 0, 2, 3)
            .unbind(0)
        )
        cos, sin = position_embeddings

        q, k = apply_rotary_pos_emb_vision(q.unsqueeze(0), k.unsqueeze(0), cos, sin)
        q = q.squeeze(0)
        k = k.squeeze(0)

        input_dtype = q.dtype
        if input_dtype == torch.float32:
            if torch.is_autocast_enabled():
                target_dtype = torch.get_autocast_gpu_dtype()
            else:
                raise ValueError("Flash attention only supports float32 input")
            q = q.to(target_dtype)
            k = k.to(target_dtype)
            v = v.to(target_dtype)

        attn_output = self.attn_func(
            q,
            k,
            v,
            cu_seqlens_q=cu_seqlens,
            cu_seqlens_k=cu_seqlens,
            max_seqlen_q=max_seqlen,
            max_seqlen_k=max_seqlen,
            causal=False,
        ).reshape(seq_length, -1)
        attn_output = self.proj(attn_output)
        return attn_output


class Qwen2_5_VLVisionBlock(nn.Module):
    def __init__(self, config: Qwen2_5_VL_Encoder_Args) -> None:
        super().__init__()
        self.norm1 = RMSNorm(
            config.hidden_size,
            config.norm_eps,
            casting_mode=config.hf_config.model_type,
        )
        self.norm2 = RMSNorm(
            config.hidden_size,
            config.norm_eps,
            casting_mode=config.hf_config.model_type,
        )
        self.attn = Qwen2_5_VLVisionAttention(
            config.hidden_size, num_heads=config.n_heads
        )
        self.mlp = Qwen2_5_VLMLP(
            config,
            bias=True,  # This is fixed to True according to the original implementation
        )

    def forward(
        self, hidden_states, cu_seqlens, position_embeddings, max_seqlen
    ) -> torch.Tensor:
        hidden_states = hidden_states + self.attn(
            self.norm1(hidden_states),
            cu_seqlens=cu_seqlens,
            position_embeddings=position_embeddings,
            max_seqlen=max_seqlen,
        )
        hidden_states = hidden_states + self.mlp(self.norm2(hidden_states))
        return hidden_states


class Qwen2_5_VisionTransformerPretrainedModel(nn.Module):
    def __init__(self, config) -> None:
        super().__init__()
        self.config = config
        self.spatial_merge_size = config.spatial_merge_size
        self.patch_size = config.patch_size
        self.fullatt_block_indexes = config.fullatt_block_indexes
        self.window_size = config.window_size
        self.spatial_merge_unit = self.spatial_merge_size * self.spatial_merge_size

        self.patch_embed = Qwen2_5_VisionPatchEmbed(
            patch_size=config.patch_size,
            temporal_patch_size=config.temporal_patch_size,
            in_channels=config.in_channels,
            embed_dim=config.hidden_size,
        )

        head_dim = config.hidden_size // config.n_heads
        self.rotary_pos_emb = Qwen2_5_VisionRotaryEmbedding(head_dim // 2)

        self.blocks = torch.nn.ModuleDict()
        for layer_id in range(config.depth):
            self.blocks[str(layer_id)] = Qwen2_5_VLVisionBlock(config)
        self.merger = Qwen2_5_VLPatchMerger(config=config)

    def rot_pos_emb(self, grid_thw):
        pos_ids = []
        for t, h, w in grid_thw:
            hpos_ids = torch.arange(h).unsqueeze(1).expand(-1, w)
            hpos_ids = hpos_ids.reshape(
                h // self.spatial_merge_size,
                self.spatial_merge_size,
                w // self.spatial_merge_size,
                self.spatial_merge_size,
            )
            hpos_ids = hpos_ids.permute(0, 2, 1, 3)
            hpos_ids = hpos_ids.flatten()

            wpos_ids = torch.arange(w).unsqueeze(0).expand(h, -1)
            wpos_ids = wpos_ids.reshape(
                h // self.spatial_merge_size,
                self.spatial_merge_size,
                w // self.spatial_merge_size,
                self.spatial_merge_size,
            )
            wpos_ids = wpos_ids.permute(0, 2, 1, 3)
            wpos_ids = wpos_ids.flatten()
            pos_ids.append(torch.stack([hpos_ids, wpos_ids], dim=-1).repeat(t, 1))
        pos_ids = torch.cat(pos_ids, dim=0)
        max_grid_size = grid_thw[:, 1:].max()
        rotary_pos_emb_full = self.rotary_pos_emb(max_grid_size)
        rotary_pos_emb = rotary_pos_emb_full[pos_ids].flatten(1)
        return rotary_pos_emb

    def get_window_index(self, grid_thw):
        window_index: list = []
        cu_window_seqlens: list = [0]
        window_index_id = 0
        vit_merger_window_size = (
            self.window_size // self.spatial_merge_size // self.patch_size
        )

        for grid_t, grid_h, grid_w in grid_thw:
            llm_grid_h, llm_grid_w = (
                grid_h // self.spatial_merge_size,
                grid_w // self.spatial_merge_size,
            )
            index = torch.arange(grid_t * llm_grid_h * llm_grid_w).reshape(
                grid_t, llm_grid_h, llm_grid_w
            )
            pad_h = vit_merger_window_size - llm_grid_h % vit_merger_window_size
            pad_w = vit_merger_window_size - llm_grid_w % vit_merger_window_size
            num_windows_h = (llm_grid_h + pad_h) // vit_merger_window_size
            num_windows_w = (llm_grid_w + pad_w) // vit_merger_window_size
            index_padded = F.pad(index, (0, pad_w, 0, pad_h), "constant", -100)
            index_padded = index_padded.reshape(
                grid_t,
                num_windows_h,
                vit_merger_window_size,
                num_windows_w,
                vit_merger_window_size,
            )
            index_padded = index_padded.permute(0, 1, 3, 2, 4).reshape(
                grid_t,
                num_windows_h * num_windows_w,
                vit_merger_window_size,
                vit_merger_window_size,
            )
            seqlens = (index_padded != -100).sum([2, 3]).reshape(-1)
            index_padded = index_padded.reshape(-1)
            index_new = index_padded[index_padded != -100]
            window_index.append(index_new + window_index_id)
            cu_seqlens_tmp = (
                seqlens.cumsum(0) * self.spatial_merge_unit + cu_window_seqlens[-1]
            )
            cu_window_seqlens.extend(cu_seqlens_tmp.tolist())
            window_index_id += (grid_t * llm_grid_h * llm_grid_w).item()
        window_index = torch.cat(window_index, dim=0)

        return window_index, cu_window_seqlens

    def forward(
        self, hidden_states: torch.Tensor, grid_thw: torch.Tensor
    ) -> torch.Tensor:
        """
        Args:
            hidden_states (`torch.Tensor` of shape `(seq_len, hidden_size)`):
                The final hidden states of the model.
            grid_thw (`torch.Tensor` of shape `(num_images_or_videos, 3)`):
                The temporal, height and width of feature shape of each image in LLM.

        Returns:
            `torch.Tensor`: hidden_states.
        """
        hidden_states = self.patch_embed(hidden_states)
        rotary_pos_emb = self.rot_pos_emb(grid_thw)
        window_index, cu_window_seqlens = self.get_window_index(grid_thw)
        cu_window_seqlens = torch.tensor(
            cu_window_seqlens,
            device=hidden_states.device,
            dtype=grid_thw.dtype if torch.jit.is_tracing() else torch.int32,
        )
        cu_window_seqlens = torch.unique_consecutive(cu_window_seqlens)

        seq_len, _ = hidden_states.size()
        hidden_states = hidden_states.reshape(
            seq_len // self.spatial_merge_unit, self.spatial_merge_unit, -1
        )
        hidden_states = hidden_states[window_index, :, :]
        hidden_states = hidden_states.reshape(seq_len, -1)
        rotary_pos_emb = rotary_pos_emb.reshape(
            seq_len // self.spatial_merge_unit, self.spatial_merge_unit, -1
        )
        rotary_pos_emb = rotary_pos_emb[window_index, :, :]
        rotary_pos_emb = rotary_pos_emb.reshape(seq_len, -1)
        emb = torch.cat((rotary_pos_emb, rotary_pos_emb), dim=-1)
        position_embeddings = (emb.cos(), emb.sin())

        cu_seqlens = torch.repeat_interleave(
            grid_thw[:, 1] * grid_thw[:, 2], grid_thw[:, 0]
        ).cumsum(
            dim=0,
            # Select dtype based on the following factors:
            #  - FA2 requires that cu_seqlens_q must have dtype int32
            #  - torch.onnx.export requires that cu_seqlens_q must have same dtype as grid_thw
            # See https://github.com/huggingface/transformers/pull/34852 for more information
            dtype=grid_thw.dtype if torch.jit.is_tracing() else torch.int32,
        )
        cu_seqlens = F.pad(cu_seqlens, (1, 0), value=0)

        with torch.no_grad():
            max_window_seqlen = (
                (cu_window_seqlens[1:] - cu_window_seqlens[:-1]).max().item()
            )
            max_seqlen = (cu_seqlens[1:] - cu_seqlens[:-1]).max().item()

        for layer_num, blk in self.blocks.items():
            if int(layer_num) in self.fullatt_block_indexes:
                cu_seqlens_now = cu_seqlens
                max_seqlen_now = max_seqlen
            else:
                cu_seqlens_now = cu_window_seqlens
                max_seqlen_now = max_window_seqlen

            if (
                hasattr(blk, "_gradient_checkpointing_enabled")
                and blk._gradient_checkpointing_enabled
            ):
                hidden_states = torch.utils.checkpoint.checkpoint(
                    blk,
                    hidden_states,
                    cu_seqlens_now,
                    position_embeddings,
                    max_seqlen_now,
                    use_reentrant=False,
                )
            else:
                hidden_states = blk(
                    hidden_states,
                    cu_seqlens=cu_seqlens_now,
                    position_embeddings=position_embeddings,
                    max_seqlen=max_seqlen_now,
                )

        hidden_states = self.merger(hidden_states)
        reverse_indices = torch.argsort(window_index)
        hidden_states = hidden_states[reverse_indices, :]

        return hidden_states

    def _get_nparams_and_flops_fn(self) -> Callable[[int], tuple[int, int]]:
        nparams = sum(p.numel() for p in self.parameters())
        nparams_embedding = sum(
            sum(p.numel() for p in m.parameters())
            for m in self.children()
            if isinstance(m, nn.Embedding)
        )

        # Reasoning behind the factor of 12 for the self-attention part of the formula:
        # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
        # 2. the flash attention does 1 more matmul recomputation in the backward
        #    but recomputation should not be counted in calculating MFU           (+0)
        # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
        # 4. we follow the convention and do not account for sparsity in causal attention
        layers, heads, head_dim = (
            len(self.blocks),
            self.config.n_heads,
            self.config.hidden_size // self.config.n_heads,
        )
        return lambda seq_len: (
            nparams,
            6 * (nparams - nparams_embedding)
            + 12 * layers * heads * head_dim * seq_len,
        )


"""
LLM part
"""


class Qwen2_5_VLRotaryEmbedding(nn.Module):
    def __init__(self, config: Qwen2_5_VL_LM_Args, device=None):
        super().__init__()
        self.config = config
        self.rope_init_fn = ROPE_INIT_FUNCTIONS[config.rope_type]
        self.reset_inv_freq(device=device)

    def reset_inv_freq(self, device: torch.device = None):
        inv_freq, self.attention_scaling = self.rope_init_fn(
            self.config.hf_config, device
        )
        if not hasattr(self, "inv_freq"):
            self.register_buffer("inv_freq", inv_freq, persistent=False)
        else:
            self.inv_freq.to(torch.float32)
            with torch.no_grad():
                self.inv_freq.data.copy_(inv_freq)

    @torch.no_grad()
    def forward(self, x, position_ids):
        if self.inv_freq.dtype != torch.float32:
            self.reset_inv_freq(device=x.device)
            assert (
                self.inv_freq.dtype == torch.float32
            ), "inv_freq dtype should be float32"
        # Core RoPE block. In contrast to other models, Qwen2_5_VL has different position ids for thw grids
        # So we expand the inv_freq to shape (3, ...)
        inv_freq_expanded = (
            self.inv_freq[None, None, :, None]
            .float()
            .expand(3, position_ids.shape[1], -1, 1)
        )
        position_ids_expanded = position_ids[
            :, :, None, :
        ].float()  # shape (3, bs, 1, positions)
        # Force float32 (see https://github.com/huggingface/transformers/pull/29285)
        device_type = x.device.type
        device_type = (
            device_type
            if isinstance(device_type, str) and device_type != "mps"
            else "cpu"
        )
        with torch.autocast(device_type=device_type, enabled=False):
            freqs = (
                inv_freq_expanded.float() @ position_ids_expanded.float()
            ).transpose(2, 3)
            emb = torch.cat((freqs, freqs), dim=-1)
            cos = emb.cos()
            sin = emb.sin()

        # Advanced RoPE types (e.g. yarn) apply a post-processing scaling factor, equivalent to scaling attention
        cos = cos * self.attention_scaling
        sin = sin * self.attention_scaling

        return cos.to(dtype=x.dtype), sin.to(dtype=x.dtype)


class Qwen2MLP(nn.Module):
    def __init__(self, config: Qwen2_5_VL_LM_Args):
        super().__init__()
        self.config = config
        self.hidden_size = config.dim
        self.intermediate_size = config.ffn_dim
        self.gate_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=False)
        self.up_proj = nn.Linear(self.hidden_size, self.intermediate_size, bias=False)
        self.down_proj = nn.Linear(self.intermediate_size, self.hidden_size, bias=False)
        self.act_mul_func = MLPActMulFunc(ACT2FN[config.hidden_act])

    def forward(self, x):
        down_proj = self.down_proj(
            self.act_mul_func(self.gate_proj(x), self.up_proj(x))
        )
        return down_proj


def apply_multimodal_rotary_pos_emb(q, k, cos, sin, mrope_section, unsqueeze_dim=2):
    """Applies Rotary Position Embedding with Multimodal Sections to the query and key tensors (https://qwenlm.github.io/blog/qwen2-vl/).

    Explanation:
        Multimodal 3D rotary position embedding is an extension to 1D rotary position embedding. The input embedding
        sequence contains vision (images / videos) embedding and text embedding or just contains text embedding. For
        vision embedding part, we apply rotary position embedding on temporal, height and width dimension seperately.
        Here we split the channel dimension to 3 chunks for the temporal, height and width rotary position embedding.
        For text embedding part, we just apply 1D rotary position embedding. The three rotary position index (temporal,
        height and width) of text embedding is always the same, so the text embedding rotary position embedding has no
        difference with modern LLMs.

    Args:
        q (`torch.Tensor`): The query tensor.
        k (`torch.Tensor`): The key tensor.
        cos (`torch.Tensor`): The cosine part of the rotary embedding.
        sin (`torch.Tensor`): The sine part of the rotary embedding.
        position_ids (`torch.Tensor`):
            The position indices of the tokens corresponding to the query and key tensors. For example, this can be
            used to pass offsetted position ids when working with a KV-cache.
        mrope_section(`List(int)`):
            Multimodal rope section is for channel dimension of temporal, height and width in rope calculation.
        unsqueeze_dim (`int`, *optional*, defaults to 1):
            The 'unsqueeze_dim' argument specifies the dimension along which to unsqueeze cos[position_ids] and
            sin[position_ids] so that they can be properly broadcasted to the dimensions of q and k. For example, note
            that cos[position_ids] and sin[position_ids] have the shape [batch_size, seq_len, head_dim]. Then, if q and
            k have the shape [batch_size, heads, seq_len, head_dim], then setting unsqueeze_dim=1 makes
            cos[position_ids] and sin[position_ids] broadcastable to the shapes of q and k. Similarly, if q and k have
            the shape [batch_size, seq_len, heads, head_dim], then set unsqueeze_dim=2.
    Returns:
        `tuple(torch.Tensor)` comprising of the query and key tensors rotated using the Rotary Position Embedding.
    """
    assert isinstance(mrope_section, list), "mrope_section must be a list"
    mrope_section = mrope_section * 2
    cos = torch.cat(
        [m[i % 3] for i, m in enumerate(cos.split(mrope_section, dim=-1))], dim=-1
    ).unsqueeze(unsqueeze_dim)
    sin = torch.cat(
        [m[i % 3] for i, m in enumerate(sin.split(mrope_section, dim=-1))], dim=-1
    ).unsqueeze(unsqueeze_dim)

    q_embed = (q * cos) + (rotate_half(q) * sin)
    k_embed = (k * cos) + (rotate_half(k) * sin)
    return q_embed, k_embed


class Qwen2_5_VLAttention(nn.Module):
    """
    Multi-head attention module.

    Args:
        model_args (Qwen2_5_VL_LM_Args): Model configuration arguments.

    Attributes:
        n_kv_heads (int): Number of key and value heads.
        n_heads (int): Number of query heads.
        n_rep (int): Number of repetitions for local heads.
        head_dim (int): Dimension size of each attention head.
        q_proj (Linear): Linear transformation for queries.
        k_proj (Linear): Linear transformation for keys.
        v_proj (Linear): Linear transformation for values.
        o_proj (Linear): Linear transformation for output.
    """

    def __init__(self, model_args: Qwen2_5_VL_LM_Args):
        super().__init__()
        self.config = model_args
        self.mrope_section = model_args.mrope_section
        assert (
            len(self.mrope_section) == 3
        ), "mrope_section must be a list of 3 integers"

        self.n_heads = model_args.n_heads
        self.n_kv_heads = model_args.n_kv_heads
        self.n_rep = self.n_heads // self.n_kv_heads
        self.head_dim = model_args.dim // model_args.n_heads
        self.attn_func = modeling_utils.flash_attn_func
        self.attn_func_varlen = modeling_utils.flash_attn_varlen_func

        self.q_proj = nn.Linear(
            model_args.dim,
            model_args.n_heads * self.head_dim,
            bias="q_proj" in model_args.biases,
        )
        self.k_proj = nn.Linear(
            model_args.dim,
            self.n_kv_heads * self.head_dim,
            bias="k_proj" in model_args.biases,
        )
        self.v_proj = nn.Linear(
            model_args.dim,
            self.n_kv_heads * self.head_dim,
            bias="v_proj" in model_args.biases,
        )
        self.o_proj = nn.Linear(
            model_args.n_heads * self.head_dim,
            model_args.dim,
            bias="o_proj" in model_args.biases,
        )

    def forward(
        self,
        x: torch.Tensor,
        position_embeddings: Tuple[torch.Tensor],
        cu_seqlens: Optional[torch.Tensor] = None,
        max_seqlen: Optional[int] = None,
    ):
        """
        Forward pass of the attention module.

        Args:
            x (torch.Tensor): Input tensor.
            position_embeddings (torch.Tensor): Position embeddings.
            cu_seqlens (Optional[torch.Tensor]): Cumulative sequence lengths for variable-length sequences.
            max_seqlen (Optional[int]): Maximum sequence length for variable-length sequences.

        Returns:
            torch.Tensor: Output tensor after attention.

        """

        bs, seqlen, _ = x.shape
        xq, xk, xv = self.q_proj(x), self.k_proj(x), self.v_proj(x)

        # Use -1 instead of `n_heads` (or `n_kv_heads`) to infer the actual
        # local heads from sizes of xq, xk, and xv as TP may have sharded them
        # after the above linear ops.
        xq = xq.view(bs, seqlen, -1, self.head_dim)
        xk = xk.view(bs, seqlen, -1, self.head_dim)
        xv = xv.view(bs, seqlen, -1, self.head_dim)

        cos, sin = position_embeddings
        xq, xk = apply_multimodal_rotary_pos_emb(xq, xk, cos, sin, self.mrope_section)

        input_dtype = xq.dtype
        if input_dtype == torch.float32:
            if torch.is_autocast_enabled():
                target_dtype = torch.get_autocast_gpu_dtype()
            else:
                raise ValueError("Flash attention only supports float32 input")
            xq = xq.to(target_dtype)
            xk = xk.to(target_dtype)
            xv = xv.to(target_dtype)

        if cu_seqlens is not None:
            xq = xq.view(seqlen, -1, self.head_dim)
            xk = xk.view(seqlen, -1, self.head_dim)
            xv = xv.view(seqlen, -1, self.head_dim)
            output = self.attn_func_varlen(
                xq,
                xk,
                xv,
                cu_seqlens,
                cu_seqlens,
                max_seqlen,
                max_seqlen,
                causal=True,
            )
        else:
            output = self.attn_func(xq, xk, xv, causal=True)
        output = output.view(bs, seqlen, -1)
        return self.o_proj(output)


class Qwen2_5_VLDecoderLayer(nn.Module):
    def __init__(self, config: Qwen2_5_VL_LM_Args, layer_idx: int):
        super().__init__()
        self.hidden_size = config.dim

        self.self_attn = Qwen2_5_VLAttention(config)
        self.mlp = Qwen2MLP(config)
        self.input_layernorm = RMSNorm(
            config.dim, config.norm_eps, casting_mode=config.hf_config.model_type
        )
        self.post_attention_layernorm = RMSNorm(
            config.dim, config.norm_eps, casting_mode=config.hf_config.model_type
        )

    def forward(
        self,
        hidden_states: torch.Tensor,
        position_embeddings: Optional[
            Tuple[torch.Tensor, torch.Tensor]
        ] = None,  # necessary, but kept here for BC
        **kwargs,
    ) -> Tuple[
        torch.FloatTensor, Optional[Tuple[torch.FloatTensor, torch.FloatTensor]]
    ]:
        """
        Args:
            hidden_states (`torch.FloatTensor`): input to the layer of shape `(batch, seq_len, embed_dim)`
            position_embeddings (`Tuple[torch.FloatTensor, torch.FloatTensor]`, *optional*):
                Tuple containing the cosine and sine positional embeddings of shape `(batch_size, seq_len, head_dim)`,
                with `head_dim` being the embedding dimension of each attention head.
            kwargs (`dict`, *optional*):
                Arbitrary kwargs to be ignored, used for FSDP and other methods that injects code
                into the model
        """

        residual = hidden_states

        # Self Attention
        hidden_states = self.self_attn(
            self.input_layernorm(hidden_states),
            position_embeddings,
            cu_seqlens=kwargs.get("cu_seqlens", None),
            max_seqlen=kwargs.get("max_seqlen", None),
        )
        hidden_states = residual + hidden_states

        # Fully Connected
        residual = hidden_states
        hidden_states = self.post_attention_layernorm(hidden_states)
        hidden_states = self.mlp(hidden_states)

        return residual + hidden_states


class Qwen2_5_VLModel(nn.Module):
    def __init__(self, config: Qwen2_5_VL_LM_Args):
        super().__init__()
        self.config = config
        self.vocab_size = config.vocab_size
        self.embed_tokens = nn.Embedding(config.vocab_size, config.dim)
        self.layers = torch.nn.ModuleDict()
        for layer_id in range(config.n_layers):
            self.layers[str(layer_id)] = Qwen2_5_VLDecoderLayer(config, layer_id)
        self.norm = RMSNorm(
            config.dim, config.norm_eps, casting_mode=config.hf_config.model_type
        )
        self.rotary_emb = Qwen2_5_VLRotaryEmbedding(config=config)
        self.lm_head = nn.Linear(config.dim, config.vocab_size, bias=False)
        self.identity_layer = IdentityLayer()

    def forward(
        self,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        interested_tokens: Optional[torch.BoolTensor] = None,
        **kwargs,  # Additional arguments for compatibility
    ):
        h = self.identity_layer(inputs_embeds)

        position_embeddings = self.rotary_emb(h, position_ids)

        if "valid_input_len" in kwargs:
            valid_input_len = kwargs["valid_input_len"]
            updated_kwargs = pack_sequences_for_inputs(
                inputs_embeds,
                valid_input_len,
                list(position_embeddings),
                interested_tokens,
                inputs_seq_dim=1,
                inputs_batch_dim=0,
                position_ids_seq_dim=2,
                position_ids_batch_dim=1,
                interested_tokens_seq_dim=1,
                interested_tokens_batch_dim=0,
                padding_mask=kwargs.get("padding_mask", None),
                cp_mesh=kwargs.get("cp_mesh", None),
            )
            position_embeddings = tuple(updated_kwargs.pop("position_ids"))
            interested_tokens = updated_kwargs.pop("interested_tokens")
            h = updated_kwargs.pop("inputs")
            h = self.identity_layer(h)
            kwargs.update(updated_kwargs)

        for layer in self.layers.values():
            if (
                hasattr(layer, "_gradient_checkpointing_enabled")
                and layer._gradient_checkpointing_enabled
            ):
                h = torch.utils.checkpoint.checkpoint(
                    layer,
                    h,
                    position_embeddings,
                    **kwargs,
                    use_reentrant=False,
                )
            else:
                h = layer(h, position_embeddings=position_embeddings, **kwargs)

        # Add `if` check just in case `pp` is enabled
        if self.norm is not None:
            if interested_tokens is not None:
                assert not isinstance(
                    h, torch.distributed.tensor.DTensor
                ), "interested_tokens must be a local tensor"
                h = h[interested_tokens]
            assert self.lm_head is not None, "lm_head must be provided in last stage"
            h = self.lm_head(self.norm(h))
        return h

    @cached_property
    def _get_nparams_and_flops_fn(self) -> Callable[[int], tuple[int, int]]:
        nparams = sum(p.numel() for p in self.parameters())
        nparams_embedding = sum(
            sum(p.numel() for p in m.parameters())
            for m in self.children()
            if isinstance(m, nn.Embedding)
        )

        # Reasoning behind the factor of 12 for the self-attention part of the formula:
        # 1. each self-attention has 2 matmul in the forward and 4 in the backward (6)
        # 2. the flash attention does 1 more matmul recomputation in the backward
        #    but recomputation should not be counted in calculating MFU           (+0)
        # 3. each matmul performs 1 multiplication and 1 addition                 (*2)
        # 4. we follow the convention and do not account for sparsity in causal attention
        layers, heads, head_dim = (
            len(self.layers),
            self.config.n_heads,
            self.config.dim // self.config.n_heads,
        )
        return lambda seq_len: (
            nparams,
            6 * (nparams - nparams_embedding)
            + 12 * layers * heads * head_dim * seq_len,
        )


@ModelRegistry.register(
    QwenVL25WeightMapper, default_data_packer_cls=Qwen2_5_VLM_DataPacker
)
class Qwen2_5_VLConditionalModel(BaseModel):
    def __init__(self, config: Qwen2_5_VL_LM_Args):
        super().__init__(config.hf_config)
        self.config = config
        self.visual = Qwen2_5_VisionTransformerPretrainedModel(config.encoder_args)
        self.model = Qwen2_5_VLModel(config.lm_args)
        self.vocab_size = config.lm_args.vocab_size

    def _process_vision_embeddings(
        self, inputs_embeds, input_ids, pixel_values, grid_thw, pad_token_id
    ):
        """Helper function to process vision embeddings (images or videos)"""
        n_tokens = (input_ids == pad_token_id).sum().item()
        if n_tokens > 0:
            vision_embeds = self.visual(pixel_values, grid_thw=grid_thw)
            assert (
                vision_embeds.shape[0] == n_tokens
            ), "vision_embeds.shape[0] must be equal to n_tokens"
            mask = input_ids == pad_token_id
            mask_unsqueezed = mask.unsqueeze(-1)
            mask_expanded = mask_unsqueezed.expand_as(inputs_embeds)
            vision_mask = mask_expanded.to(inputs_embeds.device)

            vision_embeds = vision_embeds.to(inputs_embeds.device, inputs_embeds.dtype)
            inputs_embeds = inputs_embeds.masked_scatter(vision_mask, vision_embeds)
        return inputs_embeds

    def forward(
        self,
        input_ids: Optional[torch.LongTensor] = None,
        pixel_values: Optional[torch.Tensor] = None,
        pixel_values_videos: Optional[torch.Tensor] = None,
        image_grid_thw: Optional[torch.LongTensor] = None,
        video_grid_thw: Optional[torch.LongTensor] = None,
        position_ids: Optional[torch.Tensor] = None,
        **kwargs,
    ):
        if self.model.embed_tokens is not None:
            assert input_ids.dtype in [
                torch.int32,
                torch.int64,
            ], "input_ids must be of type int32 or int64"
            inputs_embeds = self.model.embed_tokens(input_ids)
            n_image_tokens = (input_ids == self.image_token_id).sum().item()
            n_video_tokens = (input_ids == self.video_token_id).sum().item()

            # print(f"inputs_embeds: {inputs_embeds.shape}, input_ids: {input_ids.shape}, n_image_tokens: {n_image_tokens}, n_video_tokens: {n_video_tokens}")
            # get vision embeddings as tokens for next phase
            if n_image_tokens > 0:
                assert (
                    image_grid_thw is not None
                ), "image_grid_thw must be provided if there are image tokens"
                inputs_embeds = self._process_vision_embeddings(
                    inputs_embeds,
                    input_ids,
                    pixel_values,
                    image_grid_thw,
                    self.image_token_id,
                )

            if n_video_tokens > 0:
                assert (
                    video_grid_thw is not None
                ), "video_grid_thw must be provided if there are video tokens"
                inputs_embeds = self._process_vision_embeddings(
                    inputs_embeds,
                    input_ids,
                    pixel_values_videos,
                    video_grid_thw,
                    self.video_token_id,
                )
        else:
            assert (
                input_ids.is_floating_point()
            ), "input of pipeline stage > 0 must be of floating point type"
            inputs_embeds = input_ids

        # For GRPO, we can pass in the logprob_masks to the model
        # to avoid computing the logits which are not needed for the model
        outputs = self.model(
            inputs_embeds=inputs_embeds,
            # Permute back to [3, batch_size, seq_len] for Pipeline Parallelism micro batch
            position_ids=position_ids.permute(1, 0, 2).contiguous(),
            interested_tokens=kwargs.pop("interested_tokens", None),
            **kwargs,  # Additional arguments for compatibility
        )
        return outputs

    @property
    def image_token_id(self):
        return self.config.hf_config.image_token_id

    @property
    def video_token_id(self):
        return self.config.hf_config.video_token_id

    def get_position_ids(self, **kwargs) -> Tuple[torch.Tensor, torch.Tensor, int]:
        seq_dim_idx = 2
        assert "position_ids" in kwargs, "position_ids must be provided"
        return (
            kwargs["position_ids"].permute(1, 0, 2).contiguous(),
            kwargs["input_ids"],
            seq_dim_idx,
        )

    def post_to_empty_hook(self, cosmos_config: CosmosConfig):
        self.model.rotary_emb.to(torch.cuda.current_device())
        self.model.rotary_emb.reset_inv_freq()
        if self.visual is not None:
            self.visual.rotary_pos_emb.to(torch.cuda.current_device())
            self.visual.rotary_pos_emb.reset_inv_freq()

    def apply_pipeline_split(self, pp_rank, pp_size):
        """
        Apply pipeline split to the model.
        This typically involves splitting the model into multiple stages,
        and moving each stage to a different device.
        """
        assert pp_size > 1
        is_first = pp_rank == 0
        is_last = pp_rank == pp_size - 1

        # Compute the layers belonging to this stage
        n_layers = len(self.model.layers)
        layers_per_stage = n_layers // pp_size

        if not is_first:
            self.model.embed_tokens = None
            self.visual = None
        if not is_last:
            self.model.lm_head = None
            self.model.norm = None

        local_layers = torch.nn.ModuleDict()
        for i in range(
            pp_rank * layers_per_stage,
            ((pp_rank + 1) * layers_per_stage) if not is_last else n_layers,
        ):
            local_layers[str(i)] = self.model.layers[str(i)]

        # Reset the layers for pipeline splitting
        self.model.layers = local_layers

    @classmethod
    def from_model_args(
        cls, model_args: Qwen2_5_VL_Args
    ) -> "Qwen2_5_VLConditionalModel":
        """
        Initialize a GPT model from a GPTArgs object.

        Args:
            model_args (GPTArgs): Model configuration arguments.

        Returns:
            Qwen2_5_VLConditionalModel: Qwen2_5_VLConditionalModel model.

        """
        return cls(model_args)

    def load_hf_weights(
        self,
        model_name_or_path: str,
        parallel_dims: ParallelDims,
        device: torch.device,
        revision: Optional[str] = None,
    ):
        """
        Load weights from a HuggingFace model.

        Args:
            model_path (str): Path to the HuggingFace model.
            parallel_dims (ParallelDims): Parallel dimensions definition.
        """
        # Load all safetensors from `model_path`
        model_type = retry(AutoConfig.from_pretrained)(model_name_or_path).model_type
        model_path = resolve_model_path(model_name_or_path, revision=revision)
        safetensors_files = [
            f for f in os.listdir(model_path) if f.endswith(".safetensors")
        ]

        # Load LM weights
        # model.safetensors.index.json
        lm_state_dict = self.model.state_dict()
        lm_state_dict = {clear_weight_name(k): v for k, v in lm_state_dict.items()}
        # Rename dict to remove all `._orig_mod` in keys
        if self.visual is not None:
            visual_state_dict = self.visual.state_dict()
            visual_state_dict = {
                clear_weight_name(k): v for k, v in visual_state_dict.items()
            }
        else:
            visual_state_dict = {}

        with torch.device(self.current_device()):
            for f in safetensors_files:
                weights_of_ckpt = {}
                ckpt = safe_open(
                    os.path.join(model_path, f), framework="pt", device=str(device)
                )
                keys = ckpt.keys()
                for name in keys:
                    ckpt_tensor = ckpt.get_tensor(name)
                    weights_of_ckpt[name] = ckpt_tensor

                for name in weights_of_ckpt.keys():
                    tensor = weights_of_ckpt[name]
                    dest_name, shared_weight = convert_weight_from_hf(
                        tensor, name, model_type, parallel_dims
                    )
                    if dest_name in lm_state_dict:
                        target_tensor = lm_state_dict[dest_name]
                    elif dest_name in visual_state_dict:
                        target_tensor = visual_state_dict[dest_name]
                    elif parallel_dims.pp_enabled:
                        # logger.warning(f"Skipping weight: {dest_name} because it's not in the model due to pipeline split")
                        continue
                    else:
                        raise ValueError(f"Unsupported weight: {dest_name}")
                    is_dist_tensor = isinstance(
                        target_tensor, torch.distributed.tensor.DTensor
                    )
                    local_view = (
                        target_tensor.to_local() if is_dist_tensor else target_tensor
                    )
                    assert (
                        local_view.shape == shared_weight.shape
                    ), f"Shape mismatch: {local_view.shape} != {shared_weight.shape} for {dest_name} with original shape {target_tensor.shape}"
                    with torch.no_grad():
                        local_view.data.copy_(shared_weight)

    def separate_model_parts(self) -> List[nn.Module]:
        return [self.model, self.visual]

    @property
    def parallelize_fn(self) -> Tuple[Callable, nn.Module]:
        from cosmos_rl.policy.model.qwen2_5_vl.parallelize import parallelize

        return parallelize, self

    @staticmethod
    def supported_model_types():
        return ["qwen2_5_vl"]

    @classmethod
    def from_pretrained(
        cls,
        hf_config: AutoConfig,
        model_name_or_path: str,
        max_position_embeddings: Optional[int] = None,
    ) -> "Qwen2_5_VLConditionalModel":
        """
        Initialize a GPT model from a pretrained model.

        Args:
            model_name_or_path (str): Model name or path to the pretrained model.

        Returns:
            Qwen2_5_VLConditionalModel: Qwen2_5_VLConditionalModel model.

        """
        if hf_config.model_type not in cls.supported_model_types():
            raise ValueError(f"Unsupported model type: {hf_config.model_type}")

        if max_position_embeddings is None:
            max_position_embeddings = hf_config.max_position_embeddings
        else:
            hf_config.max_position_embeddings = max_position_embeddings

        vocab_size = sync_model_vocab(model_name_or_path)

        rope_scaling = {}
        if hasattr(hf_config, "rope_scaling"):
            rope_scaling = hf_config.rope_scaling or {}
        rope_type = rope_scaling.get("rope_type", rope_scaling.get("type", "default"))

        bias_list = ["q_proj", "k_proj", "v_proj"]

        lm_args = Qwen2_5_VL_LM_Args(
            mrope_section=hf_config.rope_scaling["mrope_section"],
            dim=hf_config.hidden_size,
            ffn_dim=hf_config.intermediate_size,
            n_layers=hf_config.num_hidden_layers,
            n_heads=hf_config.num_attention_heads,
            n_kv_heads=hf_config.num_key_value_heads,
            vocab_size=vocab_size,
            max_seq_len=max_position_embeddings,
            rope_theta=hf_config.rope_theta,
            norm_type="rmsnorm",
            hidden_act=hf_config.hidden_act,
            norm_eps=hf_config.rms_norm_eps,
            rope_type=rope_type,
            biases=bias_list,
            hf_config=hf_config,
        )

        encoder_args = Qwen2_5_VL_Encoder_Args(
            depth=hf_config.vision_config.depth,
            hidden_size=hf_config.vision_config.hidden_size,
            hidden_act=hf_config.vision_config.hidden_act,
            intermediate_size=hf_config.vision_config.intermediate_size,
            n_heads=hf_config.vision_config.num_heads,
            in_channels=hf_config.vision_config.in_chans,
            patch_size=hf_config.vision_config.patch_size,
            spatial_merge_size=hf_config.vision_config.spatial_merge_size,
            temporal_patch_size=hf_config.vision_config.temporal_patch_size,
            tokens_per_second=hf_config.vision_config.tokens_per_second,
            window_size=hf_config.vision_config.window_size,
            fullatt_block_indexes=hf_config.vision_config.fullatt_block_indexes,
            out_hidden_size=hf_config.vision_config.out_hidden_size,
            norm_type="rmsnorm",
            norm_eps=hf_config.rms_norm_eps,
            hf_config=hf_config.vision_config,
        )
        args = Qwen2_5_VL_Args(
            lm_args=lm_args,
            encoder_args=encoder_args,
            hf_config=hf_config,
        )
        return cls.from_model_args(args)

    def get_nparams_and_flops(self, seq_len: int) -> tuple[int, int]:
        n_params = 0
        n_flops = 0
        if self.visual is not None:
            n_params, n_flops = self.visual._get_nparams_and_flops_fn()(seq_len)
        if self.model is not None:
            lm_n_params, lm_n_flops = self.model._get_nparams_and_flops_fn(seq_len)
            n_params += lm_n_params
            n_flops += lm_n_flops
        return n_params, n_flops

    @classmethod
    def fqn_filter_for_fp8(cls) -> List[str]:
        llm = [
            "lm_head",
        ]
        visual = [
            "visual",
        ]  # Filter Linear in visual out, they will corrupt the FP8 Linear.
        return llm + visual

    def check_cp_compatible(self, cp_size: int, tp_size: int):
        visual_n_heads = self.config.encoder_args.n_heads
        llm_n_heads = self.config.lm_args.n_heads
        cp_compatible = (
            visual_n_heads % (cp_size * tp_size) == 0
            and llm_n_heads % (cp_size * tp_size) == 0
        )
        if not cp_compatible:
            raise ValueError(
                f"Model is not compatible with cp parallelism, model's visual_n_heads={visual_n_heads} or llm_n_heads={llm_n_heads} is not divisible by cp size({cp_size}) * tp_size({tp_size}) = {cp_size * tp_size}"
            )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.distributed as dist
import torch.distributed._symmetric_memory as symm_mem
import triton
import triton.language as tl

from .triton_barrier import blockwise_barrier
from .triton_utils import sync_threads


@triton.jit
def _exchange_row_offsets(
    split_sizes_ptrs,
    rank: tl.constexpr,
    world_size: tl.constexpr,
    BLOCKS_PER_REMOTE_RANK: tl.constexpr,
):
    remote_rank = tl.program_id(0) // BLOCKS_PER_REMOTE_RANK

    # split_sizes_ptr for all ranks
    # All these vector stacks into split_sizes_matrix
    split_sizes_ptrs = split_sizes_ptrs.to(tl.pointer_type(tl.uint64))

    # split_sizes_matrix[remote_rank, :]
    input_split_sizes_ptr = tl.load(split_sizes_ptrs + remote_rank).to(
        tl.pointer_type(tl.int64)
    )

    offsets_ = tl.arange(0, world_size)
    previous_split_sizes = tl.load(
        input_split_sizes_ptr + offsets_, mask=offsets_ <= rank, other=0
    )

    num_rows = tl.load(input_split_sizes_ptr + rank)
    input_row_offset = tl.sum(previous_split_sizes) - num_rows

    # split_sizes_matrix[:, rank]
    output_split_sizes_ptrs = (
        tl.load(split_sizes_ptrs + offsets_).to(tl.pointer_type(tl.int64)) + rank
    )
    output_split_sizes = tl.load(
        output_split_sizes_ptrs, mask=offsets_ <= remote_rank, other=0
    )
    output_row_offset = tl.sum(output_split_sizes) - num_rows

    return input_row_offset, output_row_offset, num_rows


@triton.jit
def on_device_all_to_all_v_kernel(
    output_ptr,
    output_splits_ptr,
    input_ptrs,
    input_splits_ptr,
    signal_pad_ptrs,
    dim: tl.constexpr,  # Separate dim for easier vectorization
    rank: tl.constexpr,
    world_size: tl.constexpr,
    BLOCKS_PER_REMOTE_RANK: tl.constexpr,
    UNROLL_FACTOR: tl.constexpr,
    BLOCK_SIZE: tl.constexpr,
):
    blockwise_barrier(signal_pad_ptrs, None, rank, world_size, sem="relaxed")
    sync_threads()

    remote_rank = tl.program_id(0) // BLOCKS_PER_REMOTE_RANK
    block_offset = tl.program_id(0) % BLOCKS_PER_REMOTE_RANK

    input_row_offset, output_row_offset, num_rows = _exchange_row_offsets(
        input_splits_ptr, rank, world_size, BLOCKS_PER_REMOTE_RANK
    )

    output_splits_ptr = output_splits_ptr.to(tl.pointer_type(tl.uint64))
    if block_offset == 0:
        # Update output_splits
        tl.store(output_splits_ptr + remote_rank, num_rows)

    input_ptr = (
        tl.load(input_ptrs.to(tl.pointer_type(tl.uint64)) + remote_rank).to(
            tl.pointer_type(tl.bfloat16)
        )
        + input_row_offset * dim
    )
    output_ptr = output_ptr + output_row_offset * dim

    outer_loop_step = BLOCK_SIZE * UNROLL_FACTOR
    outer_loop_iters_per_rank = tl.cdiv(
        tl.cdiv(num_rows * dim, outer_loop_step), BLOCKS_PER_REMOTE_RANK
    )
    numel_per_rank = outer_loop_step * outer_loop_iters_per_rank
    offset = numel_per_rank * block_offset
    end = tl.minimum(numel_per_rank * (block_offset + 1), num_rows * dim)

    unroll_region_size = (end - offset) // outer_loop_step * outer_loop_step
    for i in tl.range(offset, offset + unroll_region_size, outer_loop_step):
        for j in tl.range(
            i,
            i + outer_loop_step,
            BLOCK_SIZE,
            loop_unroll_factor=UNROLL_FACTOR,
        ):
            offsets = j + tl.arange(0, BLOCK_SIZE)
            data = tl.load(input_ptr + offsets)
            tl.store(output_ptr + offsets, data)

    offset += unroll_region_size
    while offset < end:
        offsets = offset + tl.arange(0, BLOCK_SIZE)
        mask = offsets < num_rows * dim
        data = tl.load(input_ptr + offsets, mask=mask)
        tl.store(output_ptr + offsets, data, mask=mask)
        offset += BLOCK_SIZE

    sync_threads()
    blockwise_barrier(signal_pad_ptrs, None, rank, world_size, sem="relaxed")
    return


def _on_device_all_to_all_v(
    output: torch.Tensor,
    output_splits: torch.Tensor,
    input: torch.Tensor,
    input_splits: torch.Tensor,
    group: dist.ProcessGroup = dist.group.WORLD,
    BLOCKS_PER_REMOTE_RANK=8,
    UNROLL_FACTOR: int = 8,
    BLOCK_SIZE: int = 16384,
):
    assert output.dim() == 2, f"{output.shape}"
    assert input.dim() == 2, f"{input.shape}"
    assert output.shape[1] == input.shape[1]

    dim = output.shape[1]
    input_hdl = symm_mem.rendezvous(input, group=group)
    input_splits_hdl = symm_mem.rendezvous(input_splits, group=group)

    num_blocks = input_hdl.world_size * BLOCKS_PER_REMOTE_RANK
    on_device_all_to_all_v_kernel[(num_blocks, 1, 1)](
        output,
        output_splits,
        input_hdl.buffer_ptrs_dev,
        input_splits_hdl.buffer_ptrs_dev,
        input_hdl.signal_pad_ptrs_dev,
        dim=dim,
        rank=input_hdl.rank,
        world_size=input_hdl.world_size,
        BLOCKS_PER_REMOTE_RANK=BLOCKS_PER_REMOTE_RANK,
        UNROLL_FACTOR=UNROLL_FACTOR,
        BLOCK_SIZE=BLOCK_SIZE,
        num_warps=16,
    )
    # log_triton_kernel(kernel)
    return output


class OnDeviceAllToAllV(torch.autograd.Function):
    # A symmetric memory holding the grad_output during backward
    grad_output_buf = None
    # A symmetric memory for exchanges split sizes during both forward and backward
    splits_buf = None
    # Maximum output length (need to be set before use of OnDeviceAllToAllV)
    max_output_len = None

    @staticmethod
    def forward(
        ctx,
        input: torch.Tensor,
        input_splits: torch.Tensor,
        group: dist.ProcessGroup = dist.group.WORLD,
    ):
        """
        Args:
            input: input tensor with data for all ranks concatenated.
            input_splits: input splits of shape (group.world_size,)
            group: process group to scope the collective.
        """
        # Initialize input splits buffer (one time only)
        if OnDeviceAllToAllV.splits_buf is None:
            OnDeviceAllToAllV.splits_buf = symm_mem.empty(
                *input_splits.shape,
                dtype=input_splits.dtype,
                device=input_splits.device,
            )

        if OnDeviceAllToAllV.max_output_len is None:
            raise RuntimeError(
                "Please set max output length via `OnDeviceAllToAllV.max_output_len = ...`"
            )

        # Allocate output buffer
        output = input.new_empty(OnDeviceAllToAllV.max_output_len, *input.shape[1:])
        # Allocate output splits tensor
        output_splits = torch.empty_like(input_splits)
        # Copy input splits to the buffer
        OnDeviceAllToAllV.splits_buf.copy_(input_splits)

        # Shuffle input to output
        _on_device_all_to_all_v(
            output, output_splits, input, OnDeviceAllToAllV.splits_buf, group=group
        )

        # Output splits in forward is the input splits in backward
        ctx.save_for_backward(output_splits)
        ctx.group = group
        ctx.input_shape = input.shape
        return output, output_splits

    @staticmethod
    def backward(ctx, grad_output, grad_splits):
        """
        Backward is implemented as a shuffle of the output's gradients to the input.
        Args:
            `grad_output`: output's gradients passed from the downstream.
            `grad_splits`: unused.
        """

        # Initialize grad_output buffer (one time only)
        if OnDeviceAllToAllV.grad_output_buf is None:
            assert (
                OnDeviceAllToAllV.max_output_len is not None
            ), "`max_output_len` not set"
            OnDeviceAllToAllV.grad_output_buf = symm_mem.empty(
                OnDeviceAllToAllV.max_output_len,
                *grad_output.shape[1:],
                dtype=grad_output.dtype,
                device=grad_output.device,
            )

        # TODO: is there a way to tell autograd to feed grad_output directly to
        # our symm_mem buffer?
        OnDeviceAllToAllV.grad_output_buf.narrow(0, 0, grad_output.shape[0]).copy_(
            grad_output
        )

        # Size info
        (grad_output_splits,) = ctx.saved_tensors
        OnDeviceAllToAllV.splits_buf.copy_(grad_output_splits)
        grad_input_splits = torch.empty_like(grad_output_splits)  # unused
        grad_input = grad_output.new_empty(*ctx.input_shape)

        # Shuffle gradients back to the input
        _on_device_all_to_all_v(
            grad_input,
            grad_input_splits,
            OnDeviceAllToAllV.grad_output_buf,
            OnDeviceAllToAllV.splits_buf,
            group=ctx.group,
        )
        return grad_input, None, None


# Alias
on_device_all_to_all_v = OnDeviceAllToAllV.apply



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pydantic import BaseModel, model_validator
from typing import List, Dict, Any, Optional


MESH_NAMES: List[str] = ["pp", "dp_shard", "cp", "tp"]


class Role:
    POLICY = "POLICY"
    ROLLOUT = "ROLLOUT"
    REFERENCE = "REFERENCE"
    ALL = [POLICY, ROLLOUT, REFERENCE]


class ErrorResponse(BaseModel):
    object: str = "error"
    message: str
    code: int


class HandshakeInitiatorRequest(BaseModel):
    unique_pair_name: str
    handle_base64: str


class HandshakeAcceptorRequest(BaseModel):
    unique_pair_name: str


class NcclStoreClearRequest(BaseModel):
    unique_pair_name: str


class TrainAckRequest(BaseModel):
    replica_name: str
    weight_step: int
    total_steps: int
    # For profiling
    profile_finished: bool = False
    # For logger report data
    report_data: Dict[str, Any] = {}


class WeightReadyRequest(BaseModel):
    replica_name: str


class ValidationReportRequest(BaseModel):
    src_replica_name: str
    validation_step: int
    prompt_idxs: List[int]
    payloads: List[Any]
    completions: List[List[str]]
    is_end: bool = False
    reference_answer: Optional[str] = None


class RolloutRequest(BaseModel):
    src_replica_name: str
    prompt_idxs: List[int]
    payloads: List[Any]
    completions: List[List[str]]
    is_end: bool = False
    reference_answer: Optional[str] = None


class UnregisterRequest(BaseModel):
    replica_name: str


class HeartbeatRequest(BaseModel):
    replica_name: str


class SetProfileRequest(HeartbeatRequest):
    active_steps: int = 1
    rank_filter: List[int] = [
        0,
    ]
    record_shape: bool = False
    profile_memory: bool = False
    with_stack: bool = False
    with_modules: bool = False


class SetTracePathRequest(HeartbeatRequest):
    trace_path: str
    global_rank: int


class SetShardInfosRequest(BaseModel):
    shard_infos: List[Dict[str, Any]]
    param_groups: List[List[str]]
    sorted_params: List[List[str]]


class GetShardSendRecvInstsRequest(BaseModel):
    rank: int


class RegisterRequest(BaseModel):
    replica_name: str
    role: str
    mesh_names: List[str]
    global_rank: int
    host_ip: str
    host_name: str
    ranks: List[int]
    group_size: List[int]

    @model_validator(mode="after")
    def validate_mesh_names(self):
        assert (
            len(self.mesh_names) == len(self.ranks) == len(self.group_size)
        ), "mesh_names, ranks, and group_size must have the same length"
        assert set(self.mesh_names) <= set(
            MESH_NAMES
        ), "mesh_names must be a subset of MESH_NAMES"
        assert (
            self.role in Role.ALL
        ), "role must be one of POLICY, ROLLOUT, or REFERENCE"
        assert (
            self.replica_name is not None and len(self.replica_name) > 0
        ), "replica_name must be a non-empty string being consistent within a replica"

        new_mesh_names = []
        new_ranks = []
        new_group_size = []
        original_map = {}

        # Rebuild the map with deterministic order
        for mesh_name, rank, group_size in zip(
            self.mesh_names, self.ranks, self.group_size
        ):
            original_map[mesh_name] = (rank, group_size)

        for mesh_name in MESH_NAMES:
            new_mesh_names.append(mesh_name)
            new_ranks.append(original_map[mesh_name][0])
            new_group_size.append(original_map[mesh_name][1])

        self.mesh_names = new_mesh_names
        self.ranks = new_ranks
        self.group_size = new_group_size
        return self


class NcclErrRequest(BaseModel):
    replica_name: str
    error: str



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import time
import threading
import requests
import msgpack
from functools import partial
from typing import List, Optional, Callable, NamedTuple
import torch.distributed as dist

import tensorrt_llm
from tensorrt_llm._torch.pyexecutor.py_executor import PyExecutor
from tensorrt_llm.executor.ipc import ZeroMqQueue as IpcQueue
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.rollout import TRTLLMRolloutWorkerBase
from cosmos_rl.rollout import State
from cosmos_rl.utils.parallelism_map import (
    ParallelTopoMapperGroup,
    WeightSyncInstructionsGroup,
)
from cosmos_rl.utils.util import list_to_b64, b64_to_list
from cosmos_rl.dispatcher.command import (
    BuildMeshCommand,
    PolicyToRolloutUnicastCommand,
    RolloutToRolloutBroadcastCommand,
    Command,
)
from cosmos_rl.utils.pynccl import (
    create_nccl_uid,
    create_nccl_comm,
    nccl_broadcast,
    nccl_recv,
    nccl_group_start,
    nccl_group_end,
)
from cosmos_rl.utils.api_suffix import (
    COSMOS_API_NCCL_COMM_INITIATOR_SUFFIX,
    COSMOS_API_NCCL_COMM_ACCEPTOR_SUFFIX,
    COSMOS_API_ROLLOUT_SHARD_INFOS_SUFFIX,
    COSMOS_API_ROLLOUT_SHARD_RECV_INSTS_SUFFIX,
)

from cosmos_rl.policy.model import ModelRegistry, WeightMapper
from cosmos_rl.utils import constant
from cosmos_rl.utils.logging import logger
import cosmos_rl.utils.distributed as dist_util
from cosmos_rl.utils.network_util import make_request_with_retry
import cosmos_rl.utils.util as util
from cosmos_rl.rollout.trtllm_rollout.trtllm_common import (
    ValidationInstruction,
    ShutdownInstruction,
)

trtllm_version = tensorrt_llm.__version__
logger.info(f"[Rollout] Using trtllm version: {trtllm_version}")
if trtllm_version == "1.0.0rc6":
    from cosmos_rl.rollout.trtllm_rollout.exec_loops.trtllm_1_0_0_rc6 import (
        cosmos_patched_executor_loop,
    )
else:
    raise NotImplementedError(f"Unsupported trtllm version: {trtllm_version}")

from transformers import AutoConfig

from queue import Queue

"""
1. Extend PyExecutor to support Cosmos-specific features.
2. Initialize distributed environment.
"""


class CosmosWorkerCommIpcAddrs(NamedTuple):
    replica_name_queue_addr: tuple[str, Optional[bytes]]
    weight_sync_queue_addr: tuple[str, Optional[bytes]]


class TrtLLMRolloutWorker(TRTLLMRolloutWorkerBase):
    # Trtllm will init worker twice, only for the second time, cosmos-grpo should begin to work.
    init_count = 0
    ready = False

    def __init__(self, *args, **kwargs) -> None:
        cosmos_config = kwargs.pop("cosmos_config")
        self.cosmos_ipc_queues: CosmosWorkerCommIpcAddrs = kwargs.pop(
            "cosmos_ipc_queues"
        )
        self.cosmos_replica_name_queue = IpcQueue(
            self.cosmos_ipc_queues.replica_name_queue_addr,
            is_server=False,
            name="py_executor_replica_name_queue",
        )
        self.cosmos_weight_sync_queue = IpcQueue(
            self.cosmos_ipc_queues.weight_sync_queue_addr,
            is_server=False,
            name="py_executor_weight_sync_queue",
        )

        super().__init__(*args, **kwargs)

        if TrtLLMRolloutWorker.init_count > 0:
            self.ready = True
            parallel_dims = ParallelDims.from_config(
                parallesim_config=cosmos_config.rollout.parallelism
            )
            self.parallel_dims = parallel_dims

            # build the mesh
            self.parallel_dims.build_mesh(device_type="cuda")

            self.post_init(cosmos_config, parallel_dims)

            self.cosmos_state = State()

            # CommandQueue queried from controller.
            self._command_queue: Queue[Command] = Queue()

            self.global_commnicator_idex = -1
            self.rank_in_rollout_repicas = -1

            self.policy_to_rollout_nccl_communicators = {}

            self.cosmos_batch_size = self.config.rollout.batch_size

            # For Polocy to Rollout weight mapping
            hf_config = util.retry(AutoConfig.from_pretrained)(
                self.config.policy.model_name_or_path,
                trust_remote_code=True,
            )
            self.hf_config = hf_config

            model_type = hf_config.model_type
            if not ModelRegistry.check_model_type_supported(model_type):
                logger.warning(
                    f"[Rollout] Replica can not find {model_type} in weight mapper, use {constant.COSMOS_HF_MODEL_TYPES} model type instead, with replica name: {self.replica_name}"
                )
                model_type = constant.COSMOS_HF_MODEL_TYPES
            self.weight_mapper = WeightMapper.get_weight_mapper(model_type)(hf_config)
            self.weight_mapper.setup_rollout_backend("trtllm")

            self.cosmos_model_config = hf_config
            self.enable_validation = self.config.train.enable_validation
            self.inference_stream = torch.cuda.current_stream()

            self._engine_initialized = False

            # If already registered, notify the main process.
            if hasattr(self, "replica_name"):
                # make sure all ranks went to here.
                dist.barrier()
                if self.global_rank == 0:
                    # only the rank 0 could notify the main process.
                    self.cosmos_replica_name_queue.put(self.replica_name)

        TrtLLMRolloutWorker.init_count += 1


class CosmosTRTLLMWorker(TrtLLMRolloutWorker, PyExecutor):
    """
    CosmosTRTLLMExecutor is a wrapper of PyExecutor to support Cosmos-specific features.
    P2R and R2R of cosmos-rl are implemented in this class.
    """

    def __init__(self, *args, **kwargs) -> None:
        # just call the init of PyExecutor
        super().__init__(*args, **kwargs)

    def prepare_shard_infos_for_weight_sync_insts(self):
        self.vllm_weight_inplace_view_map, grouped_recv_param_key_n_rank_list = (
            self.weight_mapper.rollout_prepare_recv(self.get_underlying_model())
        )
        self.recv_param_key_n_rank_list = []
        param_groups = []
        for group in grouped_recv_param_key_n_rank_list:
            self.recv_param_key_n_rank_list.extend(group)
            if len(group) > 1:
                param_groups.append([x[0] for x in group])
        self.recv_param_key_n_rank_list = sorted(
            self.recv_param_key_n_rank_list, key=lambda x: x[0]
        )

        local_shard_infos = ParallelTopoMapperGroup(
            self.parallel_dims,
            self.cosmos_model_config,
            is_policy=False,
            underlying_model=self.get_underlying_model(),
            backend=self.backend,
            weight_mapper=self.weight_mapper,
        ).prepare_local_shard_infos(self.recv_param_key_n_rank_list, self.global_rank)

        self.all_rank_local_shard_infos = dist_util.all_gather_object_cpu(
            local_shard_infos
        )
        all_param_groups = dist_util.all_gather_object_cpu(param_groups)
        merged_groups = {}
        for r, param_groups in enumerate(all_param_groups):
            if self.parallel_dims.get_rank_in_dim("dp_cp_tp", r) != 0:
                continue
            for group in param_groups:
                group = sorted(group)
                key = tuple(group)
                if key not in merged_groups:
                    merged_groups[key] = group
        sorted_params_all_rank = dist_util.all_gather_object_cpu(
            [x[0] for x in self.recv_param_key_n_rank_list]
        )
        sorted_params_all_rank = [
            x
            for r, x in enumerate(sorted_params_all_rank)
            if self.parallel_dims.get_rank_in_dim("dp_cp_tp", r) == 0
        ]
        if self.global_rank == 0:
            body = {
                "shard_infos": self.all_rank_local_shard_infos,
                "param_groups": list(merged_groups.values()),
                "sorted_params": sorted_params_all_rank,
            }
            data = msgpack.packb(body)
            try:
                make_request_with_retry(
                    partial(
                        requests.post,
                        data=data,
                        headers={"Content-Type": "application/msgpack"},
                    ),
                    self.get_alternative_urls(COSMOS_API_ROLLOUT_SHARD_INFOS_SUFFIX),
                    max_retries=constant.COSMOS_HTTP_RETRY_CONFIG.max_retries,
                )
            except Exception as e:
                raise RuntimeError(
                    f"[Rollout] Failed in post shard infos to controller after retries {e}."
                )

    def get_underlying_model(self):
        """
        Get the underlying parallelized model in vLLM internal.
        """
        if "qwen2_5_vl" in self.hf_config.model_type:
            if not hasattr(self.model_engine.model, "visual"):
                # trick to get visual parameters also iterated.
                self.model_engine.model.visual = (
                    self.model_engine.model.mm_encoder.visual
                )
        return self.model_engine.model

    @TRTLLMRolloutWorkerBase.register_rollout_command_handler(
        BuildMeshCommand, backend="trtllm"
    )
    def build_global_mesh(self, build_mesh_command: BuildMeshCommand):
        replica_name_to_rank = build_mesh_command.replica_name_to_rank
        if self.replica_name not in replica_name_to_rank:
            raise RuntimeError(
                f"[Rollout] Replica {self.replica_name} not found in registered replicas."
            )
        self.rank_in_rollout_repicas = replica_name_to_rank[self.replica_name]

        if len(replica_name_to_rank) == 1:
            # only one rollout replica now, no need to build mesh.
            return
        # generate key for storing the NCCL group id.
        # group_0: [rank 0 in replica 0, rank 0 in replica 1, ..., rank 0 in replica n-1]
        # group_1: [rank 1 in replica 0, rank 1 in replica 1, ..., rank 1 in replica n-1]
        # ...
        # group_m-1: [rank m-1 in replica 0, rank m-1 in replica 1, ..., rank m-1 in replica n-1]
        unique_rollout_group_key = self.get_group_unique_key(replica_name_to_rank)
        logger.debug(
            f"[Rollout] Building global mesh for {self.replica_name} with unique_rollout_group_key: {unique_rollout_group_key}"
        )

        nccl_group_id = None
        if self.rank_in_rollout_repicas == 0:
            # only replica_rank == 0 have the right to generate nccl id.
            nccl_group_id = create_nccl_uid()
            base64_nccl_group_id = list_to_b64(nccl_group_id)
            try:
                make_request_with_retry(
                    partial(
                        requests.post,
                        json={
                            "unique_pair_name": unique_rollout_group_key,
                            "handle_base64": base64_nccl_group_id,
                        },
                    ),
                    self.get_alternative_urls(COSMOS_API_NCCL_COMM_INITIATOR_SUFFIX),
                    max_retries=constant.COSMOS_HTTP_RETRY_CONFIG.max_retries,
                )
            except Exception as e:
                raise RuntimeError(
                    f"[Rollout] Failed in post nccl group_id to controller after retries {e}."
                )
            logger.debug(f"[Rollout] post nccl group_id to controller: {nccl_group_id}")
        if self.rank_in_rollout_repicas != 0:
            # other replicas should query the nccl group id from controller
            # all ranks need to wait for the rollout replica 0 finished the group_id post
            # and then they can get the group_id from controller
            # all ranks not zero in replica 0 or all ranks of other replicas need to query the group_id from controller
            nccl_group_id = self.query_nccl_unique_id_from_controller(
                unique_rollout_group_key
            )
            if nccl_group_id is None:
                raise RuntimeError(
                    "[Rollout] Failed to query nccl group_id from controller!"
                )
            logger.debug(
                f"[Rollout] query nccl group_id from controller: {nccl_group_id} for global_rank: {self.global_rank}"
            )

        # update the cached communicator index
        logger.debug(
            f"[Rollout] Creating nccl communicator for global mesh: {unique_rollout_group_key}"
        )
        self.global_commnicator_idex = create_nccl_comm(
            nccl_group_id, self.rank_in_rollout_repicas, len(replica_name_to_rank)
        )
        # update the replcia_name to rank dict
        self.replica_name_to_rank = replica_name_to_rank

    def query_nccl_unique_id_from_controller(self, unique_id_key: str):
        # We don't have something like dist.barrier(), so just use while True loop to query it like synchronize.
        nccl_group_id = None
        # all ranks not zero in replica 0 or all ranks of other replicas need to query the group_id from controller
        try:
            r = make_request_with_retry(
                partial(
                    requests.post,
                    json={"unique_pair_name": unique_id_key},
                ),
                self.get_alternative_urls(COSMOS_API_NCCL_COMM_ACCEPTOR_SUFFIX),
                max_retries=constant.COSMOS_HTTP_LONG_WAIT_MAX_RETRY,
            )
        except Exception as e:
            raise RuntimeError(
                f"[Rollout] Failed in post nccl group_id to controller after retries {e}."
            )
        base64_nccl_group_id = r.json()["handle_base64"]
        nccl_group_id = b64_to_list(base64_nccl_group_id)
        return nccl_group_id

    def recv_weight_shard(
        self,
        global_rank_of_rollout: int,
        insts_group: WeightSyncInstructionsGroup,
        communicator_index: int,
        do_weight_sync_check: bool = False,
    ):
        check_inside_group = do_weight_sync_check

        total_bytes_received = 0

        all_cloned_target_tensors = []
        tensors_to_check = []

        for insts_for_per_param in insts_group.param_instructions:
            # insts_for_per_param: WeightSyncInstructionsPerParam -> inst collection for a single tensor
            insts = insts_for_per_param.instructions
            # insts: List[Tuple[int, int, Dict[int, Any]]]
            inst_dest_name = insts_for_per_param.param_name
            target_tensor = self.vllm_weight_inplace_view_map[inst_dest_name]

            if check_inside_group:
                cloned_target_tensor = target_tensor.clone()
                # clear the current view
                target_tensor.zero_()

            for inst in insts:
                p_rank = inst.policy_rank
                r_rank = inst.rollout_rank
                tensor_split_strategys = inst.slice_strategy
                assert r_rank == global_rank_of_rollout
                vllm_tensor_view = target_tensor.cosmos_slice(tensor_split_strategys)
                recv_tensor = None
                if vllm_tensor_view.is_contiguous():
                    recv_tensor = vllm_tensor_view
                else:
                    # new a temp tensor
                    recv_tensor = torch.empty_like(vllm_tensor_view).contiguous()
                logger.debug(
                    f"Recving tensor {inst_dest_name} from policy rank {p_rank} to rollout rank {r_rank}, shape {vllm_tensor_view.shape} of {target_tensor.shape}."
                )
                nccl_recv(recv_tensor, p_rank, communicator_index)
                # inplace copy
                if not vllm_tensor_view.is_contiguous():
                    all_cloned_target_tensors.append((vllm_tensor_view, recv_tensor))

                total_bytes_received += recv_tensor.numel() * recv_tensor.element_size()

            if check_inside_group:
                tensors_to_check.append(
                    (cloned_target_tensor, target_tensor, insts, inst_dest_name)
                )

        def completion_lambda(all_cloned_target_tensors, tensors_to_check):
            for view, recv_tensor in all_cloned_target_tensors:
                view.copy_(
                    recv_tensor,
                )
            all_cloned_target_tensors.clear()

            for (
                cloned_target_tensor,
                target_tensor,
                insts,
                inst_dest_name,
            ) in tensors_to_check:
                if not torch.allclose(cloned_target_tensor, target_tensor):
                    raise ValueError(
                        f"Weight sync check failed after weight sync instruction: {insts} for {inst_dest_name}."
                    )
            tensors_to_check.clear()

        return total_bytes_received, partial(
            completion_lambda, all_cloned_target_tensors, tensors_to_check
        )

    @TRTLLMRolloutWorkerBase.register_rollout_command_handler(
        PolicyToRolloutUnicastCommand, backend="trtllm"
    )
    @torch.no_grad()
    def policy_to_rollout_unicast(self, command: PolicyToRolloutUnicastCommand):
        """
        Sync the weight from policy to rollout.
        This is Policy -> Rollout replica. Will only happen between
        a pair of policy and rollout replica.
        """
        if not self._engine_initialized:
            self.prepare_shard_infos_for_weight_sync_insts()
            self._engine_initialized = True

        if command.dst_replica_name != self.replica_name:
            return

        # get the nccl_unique_id from the controller
        communicator_index = {}
        nccl_unique_id_key = command.src_replica_name + "_" + command.dst_replica_name
        if nccl_unique_id_key in self.policy_to_rollout_nccl_communicators:
            logger.debug(
                f"[Rollout] Reusing cached communicator for {nccl_unique_id_key}"
            )
            communicator_index = self.policy_to_rollout_nccl_communicators[
                nccl_unique_id_key
            ]
        else:
            logger.debug(f"[Rollout] Querying nccl group id for {nccl_unique_id_key}")
            # query the nccl group id from controller
            nccl_group_id = self.query_nccl_unique_id_from_controller(
                nccl_unique_id_key
            )
            if nccl_group_id is None:
                raise RuntimeError(
                    "[Rollout] Failed to query nccl group_id from controller!"
                )
            # create the communicator index
            # p_rank is the rank in policy, r_rank is the rank in rollout
            communicator_index = create_nccl_comm(
                nccl_group_id,
                self.global_rank + command.src_replica_size,
                self.world_size + command.src_replica_size,
            )
            # cache the communicator index
            self.policy_to_rollout_nccl_communicators[nccl_unique_id_key] = (
                communicator_index
            )
        if not hasattr(self, "policy_to_rollout_recv_insts"):
            self.policy_to_rollout_recv_insts = []
            try:
                insts_meta = make_request_with_retry(
                    partial(
                        requests.post,
                        json={
                            "rank": self.global_rank,
                        },
                    ),
                    self.get_alternative_urls(
                        COSMOS_API_ROLLOUT_SHARD_RECV_INSTS_SUFFIX
                    ),
                    max_retries=constant.COSMOS_HTTP_RETRY_CONFIG.max_retries,
                )
                insts = msgpack.unpackb(insts_meta.content, strict_map_key=False)
                self.policy_to_rollout_recv_insts = [
                    WeightSyncInstructionsGroup.from_dict(inst) for inst in insts
                ]
            except Exception as e:
                raise RuntimeError(
                    f"[Rollout] Failed in fetching rollout from policy insts from controller after retries {e}."
                )
            logger.info(
                "[Rollout] Finished policy_to_rollout_recv_insts from controller."
            )
        total_recvs = 0
        total_params = 0
        for insts_group in self.policy_to_rollout_recv_insts:
            for insts_for_per_param in insts_group.param_instructions:
                total_params += 1
                total_recvs += len(insts_for_per_param.instructions)

        copy_stream = torch.cuda.Stream()

        with torch.cuda.stream(self.inference_stream):
            logger.debug(
                f"Starting to execute {len(self.policy_to_rollout_recv_insts)}; {total_params}, {total_recvs} weight sync receives ..."
            )
            # recv the weight from policy
            st = time.time()
            total_bytes_received = 0

            pending_bytes = [0]
            pending_completions = []
            pending_groups = 0

            def flush_completions(pending_bytes, pending_completions):
                recv_ready = torch.cuda.Event()
                recv_ready.record()
                with torch.cuda.stream(copy_stream):
                    recv_ready.wait()
                    logger.debug(
                        f"Flushing {len(pending_completions)} completions, {pending_bytes[0] // 1024 // 1024}"
                    )
                    for completion in pending_completions:
                        completion()
                    pending_bytes[0] = 0
                    pending_completions.clear()

            nccl_group_start(communicator_index)

            for insts_group in self.policy_to_rollout_recv_insts:
                # insts_group: WeightSyncInstructionsGroup -> inst collection for a full weight tensor
                # handle inst group
                bytes_received, completion_fn = self.recv_weight_shard(
                    self.global_rank,
                    insts_group,
                    communicator_index,
                    command.do_weight_sync_check,
                )
                pending_bytes[0] += bytes_received
                pending_completions.append(completion_fn)
                total_bytes_received += bytes_received

                pending_groups += 1
                if pending_groups == constant.COSMOS_P2R_NCCL_GROUP_SIZE:
                    nccl_group_end(communicator_index)
                    flush_completions(pending_bytes, pending_completions)
                    nccl_group_start(communicator_index)
                    pending_groups = 0

            nccl_group_end(communicator_index)
            flush_completions(pending_bytes, pending_completions)

            with torch.cuda.stream(copy_stream):
                copy_finished = torch.cuda.Event()
                copy_finished.record()

            copy_finished.wait()
            torch.cuda.synchronize()

            time_eclapsed = time.time() - st
            logger.info(
                f"[Rollout] All {len(self.policy_to_rollout_recv_insts)} at step {command.weight_step} recv operations finished in {time_eclapsed:.3f} seconds with {total_bytes_received / (1024 * 1024)} MB received."
            )
            self.cosmos_state.set_weight_synced()

    @TRTLLMRolloutWorkerBase.register_rollout_command_handler(
        RolloutToRolloutBroadcastCommand, backend="trtllm"
    )
    def broadcast_to_all_rollout_replica(
        self, broadcast_command: RolloutToRolloutBroadcastCommand
    ) -> None:
        """
        Broadcast the weight to all other rollout replicas.
        Will only happen between Rollout Replica 0 and all other Rollout Replicas.
        """
        src_replica_name: str = broadcast_command.src_replica_name
        dst_replica_names: List[str] = broadcast_command.dst_replica_names

        if not self._engine_initialized:
            self.prepare_shard_infos_for_weight_sync_insts()
            self._engine_initialized = True

        if len(dst_replica_names) > 1:
            # Only do broadcast if there are more than one rollout replicas.
            with torch.cuda.stream(self.inference_stream):
                assert (
                    self.rank_in_rollout_repicas >= 0
                ), "[Rollout] rank in rollout replicas should be set before broadcast."
                assert (
                    len(dst_replica_names) == len(self.replica_name_to_rank)
                ), "[Rollout] The vaild dst replicas num should match the replicas num that this worker holds."

                src_rank = self.replica_name_to_rank[src_replica_name]

                for parameter in self.get_underlying_model().state_dict().values():
                    recv_tensor = parameter
                    if not parameter.is_contiguous():
                        recv_tensor = parameter.contiguous()

                    nccl_broadcast(recv_tensor, src_rank, self.global_commnicator_idex)

                    if not parameter.is_contiguous():
                        parameter.copy_(recv_tensor)

                if not self.cosmos_state.weight_synced():
                    self.cosmos_state.set_weight_synced()

        current_step = broadcast_command.weight_step
        if current_step is not None and current_step > 0:
            should_do_validation = self.config.train.enable_validation and (
                current_step % self.config.train.validation_step == 0
                or current_step == broadcast_command.total_steps
            )
            if should_do_validation:
                self.cosmos_weight_sync_queue.put(
                    ValidationInstruction(current_step, broadcast_command.total_steps)
                )

        if broadcast_command.replica_should_stop():
            # trigger the shutdown signal to main process too.
            self.cosmos_weight_sync_queue.put(ShutdownInstruction())
            self.shutdown_signal.set()

    def query_command_from_controller(self):
        """Background task to check commands from the controller"""
        while not self.shutdown_signal.is_set():
            commands = []
            try:
                # blocking request
                commands = self.redis_controller.subscribe_command(self.replica_name)
            except Exception as e:
                logger.error(
                    f"[Rollout] Failed in query commands from controller for replica {self.replica_name}\n: {str(e)}"
                )

            for instruction in commands:
                command = Command.depack(instruction)
                logger.debug(f"[Rollout] Received command: {command.command_type}")
                self._command_queue.put(command)

    def consume_command(self, cmd_pred: Optional[Callable[[Command], bool]] = None):
        current_command = None
        if self.global_rank == 0:
            if not self._command_queue.empty():
                if cmd_pred is None:
                    current_command = self._command_queue.get()
                else:
                    if cmd_pred(self._command_queue.queue[0]):
                        current_command = self._command_queue.get()
                    else:
                        # Do not go on if the command is not expected
                        current_command = None

        current_command = dist_util.broadcast_object_cpu(current_command)

        if current_command is not None:
            handler = self.get_rollout_command_handler(
                type(current_command), backend=self.backend
            )
            if handler is None:
                raise Exception(
                    f"No such command supoorted in rollout {current_command}"
                )
            try:
                handler(self, current_command)
                logger.debug(
                    f"[Rollout] Command executed: {current_command._serialize()} for rank: {self.global_rank}"
                )
            except Exception as e:
                raise RuntimeError(
                    f"[Rollout] Command execution failed for {current_command._serialize()}"
                ) from e

    """
    Below are the methods that modified from PyExecutor.
    """

    def _executor_loop(self):
        cosmos_patched_executor_loop(self)

    def start_worker(self):
        # Start PyExecutor worker thread. The worker thread will call `_executor_loop`
        super().start_worker()

        # Start query command thread of Cosmos-RL.
        if self.global_rank == 0 and self.ready:
            # create a thread to query command as a producer
            self.background_thread = threading.Thread(
                target=self.query_command_from_controller, daemon=True
            )
            self.background_thread.start()

    def handle_shutdown(self):
        if not hasattr(self, "_shutdown_handled"):
            self._shutdown_handled = True
            if hasattr(self, "shutdown_signal") and not self.shutdown_signal.is_set():
                self.shutdown_signal.set()

        if hasattr(self, "background_thread") and self.background_thread is not None:
            self.background_thread.join()
            self.background_thread = None

        if hasattr(self, "heartbeat_thread") and self.heartbeat_thread is not None:
            self.heartbeat_thread.join()
            self.heartbeat_thread = None
        # let wrapper do this instead.
        # self.unregister_from_controller()

    def shutdown(self):
        # override pyexecutor's shutdown
        super().shutdown()
        self.handle_shutdown()



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, List
import torch
from cosmos_rl.utils.ulysses import slice_input_tensor
from torch.distributed.device_mesh import DeviceMesh


def compute_valid_lengths(input_ids, pad_token_id):
    """
    Compute the valid length at each batch.
    Used for packing sequence.
    """
    # Find the last non-pad position in each row
    # Reverse along sequence dim, find first True, convert to length
    mask = input_ids != pad_token_id  # (B, T)
    flipped_mask = mask.flip(dims=[-1])  # reverse each row
    last_non_pad = flipped_mask.long().argmax(dim=-1)  # index from end
    valid_lengths = mask.size(1) - last_non_pad  # length = total - index from end
    is_all_pad = ~mask.any(dim=-1)
    valid_lengths = valid_lengths.masked_fill(is_all_pad, 0)
    return valid_lengths


def expand_mask(mask, input, pos_seq_dim=1, batch_dim=0):
    """
    Expand the mask to match the shape of the target.
    Used for matching the mask with the target tensor at shape level.
    """

    if input.ndim > mask.ndim:
        for dim in range(input.ndim):
            if dim not in [batch_dim, pos_seq_dim]:
                mask = mask.unsqueeze(dim)
    assert input.ndim == mask.ndim, "input.ndim must be not less than mask.ndim"
    mask = mask.expand(input.shape)
    return mask


def recover_dims(updated, original, seq_dim=1, batch_dim=0):
    """
    Recover the dims layout to match the original tensor.
    Used for maching the mask with the target tensor at dimension level.
    """
    view_shape = list(original.shape)
    view_shape[seq_dim] = -1
    view_shape[batch_dim] = 1
    return updated.view(*view_shape)


def generate_mask(valid_len, input, seq_dim, batch_dim):
    """
    Generate the mask according to the valid length at each batch.
    Used for extracting the valid values and packing them.
    """
    device = input.device
    valid_mask = torch.arange(input.size(seq_dim), device=device).expand(
        input.size(batch_dim), -1
    ) < valid_len.unsqueeze(1)
    return valid_mask


def pack_sequences_info_collect(
    input_ids: torch.Tensor,
    pad_token_id: int,
    label_ids: Optional[torch.Tensor] = None,
    ignore_label_id: int = -100,
    seq_len_multiple: int = 1,
):
    """
    Collect valid length at each batch information used for sequence packing.

    input_ids: the padded input ids in [batch, seq_len] to be input to the model.
    pad_token_id: the token id used for padding the input ids.
    label_ids: the padded label ids in [batch, seq_len] used for supervised fine-tuning loss calculation.
    ignore_label_id: the token id used for ignoring certain labels during loss calculation.
    seq_len_multiple: the value which the processed seq_len must be multiple of.

    Return:
    A Dictionary including tensor "valid_input_len" which is a list of actual sequence length in the batch.
    This information will be used to apply the sequence packing.
    """
    args_dict = {}
    with torch.no_grad():
        device = input_ids.device
        # Create mask: True where not padding
        valid_input_len = compute_valid_lengths(input_ids, pad_token_id)
        if label_ids is not None:
            assert (
                input_ids.shape == label_ids.shape
            ), "Input IDs and label IDs must have the same shape"
            valid_label_len = compute_valid_lengths(label_ids, ignore_label_id)
            # Ensure label length does not exceed input length
            valid_input_len = torch.max(valid_label_len, valid_input_len)
        valid_input_len = valid_input_len.tolist()
        total_valid_input_len = sum(valid_input_len)
        origin_seq_len = input_ids.size(1)
        pad_needed = (
            total_valid_input_len + seq_len_multiple - 1
        ) // seq_len_multiple * seq_len_multiple - total_valid_input_len
        padded_valid_input_len = []
        for seq_len in valid_input_len:
            padded = min(pad_needed, origin_seq_len - seq_len)
            pad_needed -= padded
            seq_len += padded
            padded_valid_input_len.append(seq_len)
        valid_input_len = torch.tensor(
            padded_valid_input_len, dtype=torch.int32, device=device
        )
        args_dict["valid_input_len"] = valid_input_len
    return args_dict


def pack_sequences_for_inputs(
    inputs_embeds: torch.Tensor,
    valid_input_len: torch.Tensor,
    position_ids_list: Optional[List[torch.Tensor]] = None,
    interested_tokens: Optional[torch.BoolTensor] = None,
    padding_mask: Optional[torch.Tensor] = None,
    cp_mesh: DeviceMesh = None,
    inputs_seq_dim: int = 1,
    inputs_batch_dim: int = 0,
    position_ids_seq_dim: int = 1,
    position_ids_batch_dim: int = 0,
    interested_tokens_seq_dim: int = 1,
    interested_tokens_batch_dim: int = 0,
    padding_mask_seq_dim: int = 1,
    padding_mask_batch_dim: int = 0,
):
    """
    Processing input tensors inside model forward function after embedding calculation.
    Packing the tensors from different batches into one.

    inputs_embeds: the input embeddings in [batch, seq_len, hidden_size].
    valid_input_len: the valid input lengths in [batch] used for packing.
    position_ids_list: a list of position ids in [batch, seq_len] for each input sequence.
    interested_tokens: a tensor of interested tokens in [batch, seq_len] for each input sequence.
    padding_mask: a tensor of padding masks in [batch, seq_len] for each input sequence.
    cp_mesh: the device mesh for model parallelism.
    inputs_seq_dim: the sequence dimension of the input tensors.
    inputs_batch_dim: the batch dimension of the input tensors.
    position_ids_seq_dim: the sequence dimension of the position ids.
    position_ids_batch_dim: the batch dimension of the position ids.
    interested_tokens_seq_dim: the sequence dimension of the interested tokens.
    interested_tokens_batch_dim: the batch dimension of the interested tokens.
    padding_mask_seq_dim: the sequence dimension of the padding mask.
    padding_mask_batch_dim: the batch dimension of the padding mask.

    Return:
    A Dictionary including all related tensors generated for model inputs after sequence packing.
    Including sequence packed "inputs", "interested_tokens", "padding_mask", and "position_ids".
    After packing, the batch dimension will be 1 and the seq_len will be the sum of all inputs lengths with some paddings in a packing way.
    Also including "cu_seqlens" for accumulated sequence length for the batchs.
    And "max_seqlen" for the max sequence length in the batch.
    These will be used by the model to do the forward and backward in a sequence packing way.
    """

    args_dict = {}
    with torch.no_grad():
        device = inputs_embeds.device
        valid_input_mask = generate_mask(
            valid_input_len, inputs_embeds, inputs_seq_dim, inputs_batch_dim
        )
        packed_inputs_embeds = recover_dims(
            inputs_embeds[valid_input_mask],
            inputs_embeds,
            inputs_seq_dim,
            inputs_batch_dim,
        )
        if cp_mesh is not None:
            packed_inputs_embeds = slice_input_tensor(
                packed_inputs_embeds, inputs_seq_dim, cp_mesh
            )
        args_dict["inputs"] = packed_inputs_embeds

        if interested_tokens is not None:
            assert interested_tokens.ndim == 2, "ndim of interested_tokens must be 2"
            interested_tokens = recover_dims(
                interested_tokens[valid_input_mask],
                interested_tokens,
                interested_tokens_seq_dim,
                interested_tokens_batch_dim,
            )
            if cp_mesh is not None:
                interested_tokens = slice_input_tensor(
                    interested_tokens, interested_tokens_seq_dim, cp_mesh
                )

        args_dict["interested_tokens"] = interested_tokens

        if padding_mask is not None:
            assert padding_mask.ndim == 2, "ndim of padding_mask must be 2"
            padding_mask = recover_dims(
                padding_mask[valid_input_mask],
                padding_mask,
                padding_mask_seq_dim,
                padding_mask_batch_dim,
            )
            if cp_mesh is not None:
                padding_mask = slice_input_tensor(
                    padding_mask, padding_mask_seq_dim, cp_mesh
                )

        args_dict["padding_mask"] = padding_mask

        position_ids_updated = []
        if position_ids_list is not None:
            for position_ids in position_ids_list:
                position_ids_mask = expand_mask(
                    valid_input_mask,
                    position_ids,
                    position_ids_seq_dim,
                    position_ids_batch_dim,
                )
                position_ids = recover_dims(
                    position_ids[position_ids_mask],
                    position_ids,
                    position_ids_seq_dim,
                    position_ids_batch_dim,
                )
                if cp_mesh is not None:
                    position_ids = slice_input_tensor(
                        position_ids, position_ids_seq_dim, cp_mesh
                    )
                position_ids_updated.append(position_ids)
        args_dict["position_ids"] = position_ids_updated

        seq_len = valid_input_len
        prefill_start_pos = torch.cumsum(seq_len, dim=0, dtype=torch.int32) - seq_len
        cu_seqlens = torch.cat(
            [
                prefill_start_pos,
                torch.tensor([torch.sum(seq_len)], dtype=torch.int32).to(device),
            ],
            dim=0,
        )
        args_dict["cu_seqlens"] = cu_seqlens
        args_dict["max_seqlen"] = max(valid_input_len.tolist())
    return args_dict


def pack_sequences_for_masks(
    valid_input_len: torch.Tensor,
    valid_label_len: torch.Tensor,
):
    """
    Generate the related masks for packed input sequences used for loss calculation.

    valid_input_len: the valid input lengths for each sequence in the batch of shape [batch].
    valid_label_len: the valid label lengths for each sequence in the batch of shape [batch].

    Return:
    A Dictionary including all related masks generated for model inputs and labels after sequence packing.
    "input_packing_mask" is the mask for extracting the useful positions from the packed inputs.
    "label_packing_mask" is the mask for extracting the useful positions from the packed labels.
    The extracted useful positions from inputs and labels will be compared to calculate the loss.
    """

    args_dict = {}
    with torch.no_grad():
        device = valid_input_len.device
        valid_input_len = valid_input_len.tolist()
        input_packing_mask = []
        for valid_len in valid_input_len:
            input_packing_mask.extend([1] * (valid_len - 1) + [0])
        input_packing_mask = torch.tensor([input_packing_mask], dtype=torch.bool).to(
            device
        )
        args_dict["input_packing_mask"] = input_packing_mask
        label_packing_mask = []
        for valid_len in valid_label_len.tolist():
            label_packing_mask.extend([0] + [1] * (valid_len - 1))
        label_packing_mask = torch.tensor([label_packing_mask], dtype=torch.bool).to(
            device
        )
        args_dict["label_packing_mask"] = label_packing_mask
    return args_dict


def pack_sequences_for_labels(
    label_ids: torch.Tensor,
    valid_label_len: torch.Tensor,
    label_ids_seq_dim: int = 1,
    label_ids_batch_dim: int = 0,
):
    """
    Generate packed label ids to pack from different batches into one used for SFT loss calculation.

    label_ids: the label ids of shape [batch, seq_len] for each sequence in the batch.
    valid_label_len: the valid label lengths of shape [batch, seq_len] for each sequence in the batch.
    label_ids_seq_dim: the sequence dimension of the label ids tensor.
    label_ids_batch_dim: the batch dimension of the label ids tensor.

    Return:
    The packed label ids after sequence packing.
    It will be used with the packed model outputs to get the loss in SFT.
    """
    with torch.no_grad():
        valid_label_mask = generate_mask(
            valid_label_len, label_ids, label_ids_seq_dim, label_ids_batch_dim
        )
        packed_label_ids = recover_dims(
            label_ids[valid_label_mask],
            label_ids,
            label_ids_seq_dim,
            label_ids_batch_dim,
        )
    return packed_label_ids


def pack_sequences_for_logprobs(
    logprob_masks: torch.Tensor,
    valid_input_len: torch.Tensor,
    advantages: Optional[torch.Tensor] = None,
    input_ids_seq_dim: int = 1,
    input_ids_batch_dim: int = 0,
):
    """
    Generate packed log probabilities and advantages to pack from different batches into one used for GRPO loss calculation.

    logprob_masks: the log probability masks of shape [batch, seq_len] for each sequence in the batch.
    valid_input_len: the valid input lengths of shape [batch] for each sequence in the batch.
    advantages: the advantages of shape [batch] for each sequence in the batch.
    input_ids_seq_dim: the sequence dimension of the input ids tensor.
    input_ids_batch_dim: the batch dimension of the input ids tensor.

    Return:
    A Dictionary including related logprob_masks and advantages after sequence packing.
    "logprob_masks" and "advantages" are the logprob_masks and advantages after sequence packing.
    They will be used to calculate the loss together with the packed model inputs and outputs in RL.
    """
    args_dict = {}
    with torch.no_grad():
        device = logprob_masks.device
        valid_input_mask = generate_mask(
            valid_input_len, logprob_masks, input_ids_seq_dim, input_ids_batch_dim
        )
        packed_logprob_masks = recover_dims(
            logprob_masks[valid_input_mask],
            logprob_masks,
            input_ids_seq_dim,
            input_ids_batch_dim,
        )
        args_dict["logprob_masks"] = packed_logprob_masks
        valid_input_len = valid_input_len.tolist()
        concatenated_advantages = [
            adv
            for idx, adv in enumerate(advantages)
            for _ in range(valid_input_len[idx])
        ]
        advantages = torch.tensor(
            [concatenated_advantages],
            dtype=torch.float32,
        ).to(device)
        args_dict["advantages"] = advantages
    return args_dict



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import torch
import time

from cuda import cudart
from tensorrt_llm._torch.pyexecutor.py_executor import BatchState
from tensorrt_llm._torch.pyexecutor.llm_request import LlmRequestState
from tensorrt_llm._torch.pyexecutor.sampler import SampleState
from tensorrt_llm.runtime.generation import CUASSERT


def cosmos_patched_executor_loop(self):
    torch.cuda.set_device(self.device_id)
    # ensure the context is created, otherwise, some MPI calls will fail.
    CUASSERT(cudart.cudaSetDevice(self.device_id))
    with self._profiler() as profile_step:
        sample_state = None
        iter_start_time = time.time()
        iter_stats = None
        while True:
            # Cosmos-RL specific code start
            if self.ready:
                self.consume_command(cmd_pred=None)
                if not self.cosmos_state.weight_synced():
                    continue  # if weight is not synced, skip the generation and while-loop until weight is synced.
            # Cosmos-RL specific code end

            profile_step()
            if self.enable_iter_perf_stats:
                iter_start_time = time.time()

            scheduled_batch, iter_stats = self._prepare_and_schedule_batch()
            if scheduled_batch is None:
                break

            self._pause_requests(scheduled_batch.paused_requests)

            finished_requests = []

            if scheduled_batch.batch_size > 0 or (
                self.enable_attention_dp and self.dist.tp_size > 1
            ):
                if self.kv_cache_transceiver:
                    # For generation requests which have completed KV cache transfer
                    self._prepare_disagg_gen_transmission_complete(scheduled_batch)

                    # Return the first token to the client
                    self._handle_first_token_response(scheduled_batch)

                self.resource_manager.prepare_resources(scheduled_batch)
                if self.drafter is not None and self.use_spec_decode:
                    self.drafter.prepare_draft_tokens(
                        scheduled_batch, self.resource_manager
                    )

                batch_outputs = self._forward_step(scheduled_batch)
                self._execute_guided_decoder(scheduled_batch, batch_outputs["logits"])

                sample_state = self._sample_async(scheduled_batch, batch_outputs)

                self._update_request_states(scheduled_batch)
                self._update_requests(sample_state)

                if self.kv_cache_transceiver:
                    ctx_transmission_reqs = self._send_disagg_ctx_cache(
                        scheduled_batch.context_requests
                    )
                    # For context only req in transmission, we reset the state since sampler might have changed it
                    for req in ctx_transmission_reqs:
                        req.state = LlmRequestState.DISAGG_CONTEXT_TRANS_IN_PROGRESS

                self._handle_canceled_requests()
                finished_requests = self._handle_responses()
                self.resource_manager.update_resources(scheduled_batch)
                if self.enable_kv_cache_events:
                    self._add_kv_cache_events()

            if self.kv_cache_transceiver and self.ctx_in_transmission_requests:
                self._terminate_ctx_finished_requests()

            if self.enable_iter_perf_stats:
                iter_stats.inflight_batching_stats.num_ctx_tokens = (
                    self.model_engine.iter_states["num_ctx_tokens"]
                )
                self._process_iter_stats(
                    finished_requests,
                    self.active_requests,
                    BatchState(
                        sample_state=SampleState(scheduled_requests=scheduled_batch),
                        iter_stats=iter_stats,
                        iter_start_time=iter_start_time,
                    ),
                )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from typing import Iterable, List, Optional, Tuple
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from cosmos_rl.policy.config import LoraConfig
from cosmos_rl.utils.logging import logger
import os


class WeightWrapper(nn.Module):
    def __init__(self, weight: nn.Parameter):
        super().__init__()
        self.weight = weight


class LoraInjectedLinear(nn.Linear):
    """
    nn.Linear with optional LoRA adapters:
      y = x W^T + b + scale * ( (dropout(x) A^T) B^T )
      where ΔW = B @ A and scale = lora_alpha / r
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        r: int = 0,
        lora_alpha: float = 1.0,
        lora_dropout: float = 0.0,
        use_rslora: bool = False,
        device=None,
        dtype=None,
    ) -> None:
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__(in_features, out_features, bias=bias, **factory_kwargs)

        self.r = int(r)
        self.lora_alpha = float(lora_alpha)
        self.lora_dropout_p = float(lora_dropout)
        self.merged = False

        assert self.r > 0, "LoRA rank must be greater than 0"
        # LoRA parameters
        self.lora_A = WeightWrapper(
            nn.Parameter(torch.empty(self.r, in_features, **factory_kwargs))
        )
        self.lora_B = WeightWrapper(
            nn.Parameter(torch.empty(out_features, self.r, **factory_kwargs))
        )
        # Init as in the LoRA paper: A ~ N(0, 0.02), B = 0 so initial ΔW=0
        nn.init.kaiming_uniform_(self.lora_A.weight, a=math.sqrt(5))
        nn.init.zeros_(self.lora_B.weight)
        self.lora_dropout = (
            nn.Dropout(self.lora_dropout_p)
            if self.lora_dropout_p > 0.0
            else nn.Identity()
        )
        if use_rslora:
            self.scaling = self.lora_alpha / math.sqrt(self.r)
        else:
            self.scaling = self.lora_alpha / self.r

    @torch.no_grad()
    def reinitialize_lora_params(self):
        nn.init.kaiming_uniform_(self.lora_A.weight, a=math.sqrt(5))
        nn.init.zeros_(self.lora_B.weight)

    @classmethod
    def from_linear(
        cls,
        base: nn.Linear,
        r: int,
        lora_alpha: float,
        lora_dropout: float,
        use_rslora: bool = False,
    ) -> "LoraInjectedLinear":
        # Create same-shaped Linear and copy weights/bias
        new = cls(
            base.in_features,
            base.out_features,
            bias=base.bias is not None,
            r=r,
            lora_alpha=lora_alpha,
            lora_dropout=lora_dropout,
            device=base.weight.device,
            dtype=base.weight.dtype,
            use_rslora=use_rslora,
        )
        with torch.no_grad():
            new.weight.copy_(base.weight)
            if base.bias is not None:
                new.bias.copy_(base.bias)
        return new

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = F.linear(x, self.weight, self.bias)
        if self.r > 0 and not self.merged:
            # (x A^T) B^T = x @ A^T @ B^T
            after_A = self.lora_dropout(x) @ self.lora_A.weight.t()  # [*, r]
            lora_out = after_A @ self.lora_B.weight.t()  # [*, out]
            out = out + self.scaling * lora_out
        return out

    @torch.no_grad()
    def merge_adapters_(self) -> None:
        if self.r == 0 or self.merged:
            return
        lora_A = (
            self.lora_A.weight.full_tensor()
            if isinstance(self.lora_A.weight, torch.distributed.tensor.DTensor)
            else self.lora_A.weight
        )
        lora_B = (
            self.lora_B.weight.full_tensor()
            if isinstance(self.lora_B.weight, torch.distributed.tensor.DTensor)
            else self.lora_B.weight
        )
        delta_w = self.scaling * (lora_B @ lora_A)  # [out, in]

        assert not isinstance(delta_w, torch.distributed.tensor.DTensor)

        if isinstance(self.weight, torch.distributed.tensor.DTensor):
            self.weight.add_(
                torch.distributed.tensor.DTensor.from_local(
                    delta_w,
                    device_mesh=self.weight.device_mesh,
                    placements=[torch.distributed.tensor.Replicate()],
                    run_check=False,
                )
            )
        else:
            self.weight.add_(delta_w)
        self.merged = True

    @torch.no_grad()
    def unmerge_adapters_(self) -> None:
        if self.r == 0 or not self.merged:
            return

        lora_A = (
            self.lora_A.weight.full_tensor()
            if isinstance(self.lora_A.weight, torch.distributed.tensor.DTensor)
            else self.lora_A.weight
        )
        lora_B = (
            self.lora_B.weight.full_tensor()
            if isinstance(self.lora_B.weight, torch.distributed.tensor.DTensor)
            else self.lora_B.weight
        )
        delta_w = self.scaling * (lora_B @ lora_A)  # [out, in]
        assert not isinstance(delta_w, torch.distributed.tensor.DTensor)
        if isinstance(self.weight, torch.distributed.tensor.DTensor):
            self.weight.sub_(
                torch.distributed.tensor.DTensor.from_local(
                    delta_w,
                    device_mesh=self.weight.device_mesh,
                    placements=[torch.distributed.tensor.Replicate()],
                    run_check=False,
                )
            )
        else:
            self.weight.sub_(delta_w)
        self.merged = False

    def lora_parameters(self) -> Iterable[nn.Parameter]:
        if self.r == 0:
            return []
        return [self.lora_A.weight, self.lora_B.weight]


def _name_matches(name: str, needles: Iterable[str]) -> bool:
    for n in needles:
        if n and n in name:
            return True
    return False


def inject_lora_adapters(
    model: nn.Module, config: LoraConfig
) -> Tuple[nn.Module, List[str]]:
    """
    Replace matching nn.Linear modules with LoraInjectedLinear.
    Returns (model, replaced_module_names).

    If a leaf module is listed in config.modules_to_save, it is treated as
    fully tunable and LoRA will NOT be injected into it.
    """
    if not config.target_modules:
        raise ValueError(
            "LoraConfig.target_modules must be set (list of substrings to match)."
        )

    replaced: List[str] = []

    match_all_linear = config.target_modules == "all-linear"

    # Normalize modules_to_save into a list of patterns (or empty list)
    modules_to_save = getattr(config, "modules_to_save", []) or []
    if isinstance(modules_to_save, str):
        modules_to_save = [modules_to_save]
    # logger.info(f"modules_to_save: {modules_to_save}")

    def _in_modules_to_save(qualified: str, child: str) -> bool:
        if not modules_to_save:
            return False
        # if 'visual' in qualified:
        #     print(f"qualified: {qualified}, child: {child}, child in modules_to_save: {child in modules_to_save}, {_name_matches(qualified, modules_to_save)}")
        # Support both exact leaf-name matches and substring/pattern matches
        return child in modules_to_save or _name_matches(qualified, modules_to_save)

    lm_head_module_name = "lm_head"
    from cosmos_rl.policy.model.hf_models import HFModel

    if isinstance(model, HFModel):
        # Must enable input require grads for the HFModel to work with LoRA.
        # Otherwise, the model will raise a RuntimeError:
        # `element 0 of tensors does not require grad and does not have a grad_fn`.
        model.model.enable_input_require_grads()
        output_layer = model.model.get_output_embeddings()
        lm_head_module_name = [
            name for name, module in model.named_modules() if module is output_layer
        ][0]

    replaced_names = []
    for module_name, module in list(model.named_modules()):
        parent: Optional[nn.Module] = _get_parent_by_qualified_name(model, module_name)
        if parent is None:
            continue

        child_name = module_name.split(".")[-1]

        # If user asked to fully-tune this module, skip injecting LoRA here.
        # (Covers Linear/Conv1d; extend tuple if you later support more types.)
        if isinstance(module, (nn.Linear, nn.Conv1d)) and _in_modules_to_save(
            module_name, child_name
        ):
            logger.info(
                f"Skipping add lora adapter to {module_name} because it is in modules_to_save (fully tunable)."
            )
            continue

        if isinstance(module, nn.Linear):
            if match_all_linear or _name_matches(module_name, config.target_modules):
                # exclude output layer if wildcard match is enabled
                if match_all_linear and (
                    child_name == lm_head_module_name
                    or _name_matches(module_name, ["lm_head"])
                ):
                    logger.info(
                        f"Skipping {module_name} because it is the output layer"
                    )
                    continue

                lora_linear = LoraInjectedLinear.from_linear(
                    base=module,
                    r=config.r,
                    lora_alpha=config.lora_alpha,
                    lora_dropout=config.lora_dropout,
                    use_rslora=config.use_rslora,
                )
                setattr(parent, child_name, lora_linear)
                replaced.append(module_name)
                replaced_names.append(module_name)

    if not replaced:
        raise RuntimeError(
            "inject_lora_adapters found no matching nn.Linear modules. "
            f"target_modules={config.target_modules}"
        )
    else:
        if os.environ.get("RANK", "0") == "0":
            logger.info(
                f"Replaced {len(replaced_names)} modules with LoRA adapters: {replaced_names}"
            )
    return model, replaced


def _get_parent_by_qualified_name(
    root: nn.Module, qualified_name: str
) -> Optional[nn.Module]:
    """
    For a qualified child name like 'transformer.h.0.attn.q_proj',
    return the parent module object.
    """
    if not qualified_name:
        return None
    parts = qualified_name.split(".")
    if len(parts) == 1:
        return root
    parent = root
    for p in parts[:-1]:
        parent = getattr(parent, p, None)
        if parent is None:
            return None
    return parent


def mark_only_lora_as_trainable(model: nn.Module, config: LoraConfig) -> None:
    """
    Freeze all non-LoRA parameters; enable grads only for LoRA parameters.
    """
    for p in model.parameters():
        p.requires_grad = False

    # Enable grads only for the modules in config.modules_to_save
    if config.modules_to_save is not None:
        for module_name, module in model.named_modules():
            if _name_matches(module_name, config.modules_to_save):
                for p in module.parameters():
                    p.requires_grad = True

    # Enable grads only for the LoRA parameters
    for m in model.modules():
        if isinstance(m, LoraInjectedLinear) and m.r > 0:
            m.lora_A.requires_grad_(True)
            m.lora_B.requires_grad_(True)


def reinitialize_lora_params(model: nn.Module) -> None:
    for m in model.modules():
        if isinstance(m, LoraInjectedLinear):
            m.reinitialize_lora_params()


@torch.no_grad()
def merge_lora_weights_(model: nn.Module) -> None:
    for m in model.modules():
        if isinstance(m, LoraInjectedLinear):
            m.merge_adapters_()


@torch.no_grad()
def unmerge_lora_weights_(model: nn.Module) -> None:
    for m in model.modules():
        if isinstance(m, LoraInjectedLinear):
            m.unmerge_adapters_()


def lora_state_dict(model: nn.Module, prefix: str = "lora") -> dict:
    """
    Return only LoRA parameters in a flat dict:
      f"{prefix}.{module_qualified_name}.lora_A" / ".lora_B"
    """
    out = {}
    for name, module in model.named_modules():
        if isinstance(module, LoraInjectedLinear) and module.r > 0:
            out[f"{prefix}.{name}.lora_A.weight"] = (
                module.lora_A.weight.detach().clone()
            )
            out[f"{prefix}.{name}.lora_B.weight"] = (
                module.lora_B.weight.detach().clone()
            )
    return out


@torch.no_grad()
def load_lora_state_dict(
    model: nn.Module, state: dict, prefix: str = "lora", strict: bool = True
) -> None:
    """
    Load only LoRA params from a state dict produced by lora_state_dict().
    Accepts keys like: f"{prefix}.{qualified.module.name}.lora_A" / ".lora_B"
    """
    missing = []

    for key, tensor in state.items():
        if not key.startswith(prefix + "."):
            continue

        # Strip "prefix."
        tail = key[len(prefix) + 1 :]  # "<qual.name>.lora_A"
        try:
            qual_name, leaf = tail.rsplit(".", 1)  # ("qual.name", "lora_A" or "lora_B")
        except ValueError:
            if strict:
                missing.append(key)
            continue

        module = _get_module_by_qualified_name(model, qual_name)
        if not isinstance(module, WeightWrapper) or not hasattr(module, leaf):
            if strict:
                missing.append(key)
            continue

        param = getattr(module, leaf)
        param.copy_(tensor.to(dtype=param.dtype, device=param.device))

    if strict and missing:
        raise KeyError(f"Missing LoRA destinations for keys: {missing}")


def _get_module_by_qualified_name(
    root: nn.Module, qualified_name: str
) -> Optional[nn.Module]:
    m = root
    for p in qualified_name.split("."):
        m = getattr(m, p, None)
        if m is None:
            return None
    return m



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Callable
from enum import Enum


class ParallelismStrategyRole(Enum):
    POLICY = 1
    ROLLOUT = 1 << 1
    ALL = POLICY | ROLLOUT

    def __or__(self, other):
        return self.value | other.value

    def __and__(self, other):
        return self.value & other.value


_PARALLELISM_STRATEGY_REGISTRY: Dict[int, Dict[str, Callable]] = {
    ParallelismStrategyRole.POLICY: {},
    ParallelismStrategyRole.ROLLOUT: {},
}


def register_parallelism_strategy(
    reg_key: str,
    *,
    allow_override: bool = False,
    role: ParallelismStrategyRole = ParallelismStrategyRole.ALL,
):
    def decorator(func: Callable) -> Callable:
        if role & ParallelismStrategyRole.POLICY:
            if (
                not allow_override
                and reg_key
                in _PARALLELISM_STRATEGY_REGISTRY[ParallelismStrategyRole.POLICY]
            ):
                raise ValueError(f"Function '{reg_key}' is already registered.")
            _PARALLELISM_STRATEGY_REGISTRY[ParallelismStrategyRole.POLICY][reg_key] = (
                func
            )
        if role & ParallelismStrategyRole.ROLLOUT:
            if (
                not allow_override
                and reg_key
                in _PARALLELISM_STRATEGY_REGISTRY[ParallelismStrategyRole.ROLLOUT]
            ):
                raise ValueError(f"Function '{reg_key}' is already registered.")
            _PARALLELISM_STRATEGY_REGISTRY[ParallelismStrategyRole.ROLLOUT][reg_key] = (
                func
            )
        return func

    return decorator


def query_parallelism_strategy(
    model_name: str, role: ParallelismStrategyRole
) -> Callable:
    return _PARALLELISM_STRATEGY_REGISTRY[role][model_name]


def get_policy_parallelism_strategy(model_name: str) -> Callable:
    return query_parallelism_strategy(model_name, ParallelismStrategyRole.POLICY)


def get_rollout_parallelism_strategy(model_name: str) -> Callable:
    return query_parallelism_strategy(model_name, ParallelismStrategyRole.ROLLOUT)



import torch.distributed as dist
from mpi4py import MPI
import os
import torch
import random
from tensorrt_llm._utils import mpi_broadcast

from cosmos_rl.utils.constant import COSMOS_HTTP_RETRY_CONFIG
from cosmos_rl.utils.logging import logger
from cosmos_rl.utils.network_util import find_available_port

OMPI_COMM_TYPE_HOST = 9

global_comm = MPI.COMM_WORLD


def set_mpi_comm(new_comm):
    """
    Set the MPI communicator to be used by the distributed package.
    """
    global global_comm
    global_comm = new_comm


def mpi_comm():
    return global_comm


local_comm = mpi_comm().Split_type(split_type=OMPI_COMM_TYPE_HOST)


local_comm = mpi_comm().Split_type(split_type=OMPI_COMM_TYPE_HOST)


def mpi_rank():
    return mpi_comm().Get_rank()


def global_mpi_rank():
    return MPI.COMM_WORLD.Get_rank()


def global_mpi_size():
    return MPI.COMM_WORLD.Get_size()


def mpi_world_size():
    return mpi_comm().Get_size()


def local_mpi_rank():
    return local_comm.Get_rank()


def local_mpi_size():
    return local_comm.Get_size()


def init_distributed_with_MPI():
    # FIXME: (lms) Support multi-nodes.
    local_rank = mpi_rank()
    global_rank = global_mpi_rank()
    world_size = global_mpi_size()

    if world_size == 1:
        return

    if dist.is_initialized():
        return

    cosmos_world_size = os.environ.get("COSMOS_WORLD_SIZE", None)
    cosmos_local_world_size = os.environ.get("COSMOS_LOCAL_WORLD_SIZE", None)

    assert cosmos_world_size is not None, "COSMOS_WORLD_SIZE is not set."
    assert cosmos_local_world_size is not None, "COSMOS_LOCAL_WORLD_SIZE is not set."

    if cosmos_world_size is not None:
        assert world_size == int(
            cosmos_world_size
        ), "COSMOS_WORLD_SIZE is not consistent with the world size of MPI."
    if cosmos_local_world_size is not None:
        assert mpi_world_size() == int(
            cosmos_local_world_size
        ), "COSMOS_LOCAL_WORLD_SIZE is not consistent with the local world size of MPI."

    rdzv_endpoint = os.environ.get("COSMOS_RDZV_ENDPOINT", None)
    assert rdzv_endpoint is not None, "COSMOS_RDZV_ENDPOINT is not set."
    rdzv_host, rdzv_port = rdzv_endpoint.split(":")

    torch.cuda.set_device(local_rank)
    max_port = 65535
    min_port = 12371

    for _ in range(COSMOS_HTTP_RETRY_CONFIG.max_retries):
        if not int(rdzv_port):
            rdzv_port = None
            if mpi_rank() == 0:
                rdzv_port = find_available_port(
                    start_port=random.randint(min_port, max_port)
                )
            rdzv_port = mpi_broadcast(rdzv_port, root=0)

        os.environ["LOCAL_RANK"] = str(local_rank)
        os.environ["RANK"] = str(global_rank)
        os.environ["WORLD_SIZE"] = str(world_size)
        os.environ["MASTER_ADDR"] = rdzv_host
        os.environ["MASTER_PORT"] = str(rdzv_port)

        init_method = f"tcp://{os.environ['MASTER_ADDR']}:{os.environ['MASTER_PORT']}"

        try:
            # We use nccl and gloo backend
            dist.init_process_group(
                "cuda:nccl,cpu:gloo",
                world_size=world_size,
                rank=local_rank,
                init_method=init_method,
            )
        except dist.DistNetworkError:
            continue
        if dist.is_initialized():
            break
    else:
        raise RuntimeError(
            f"Failed to initialize distributed environment after {COSMOS_HTTP_RETRY_CONFIG.max_retries} retries."
        )

    logger.info(
        f"[Rollout] init torch distributed environment inside trtllm worker with tcp://{rdzv_host}:{rdzv_port} in rank {local_rank}."
    )



# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import toml
import os
import time
import argparse
from torch.utils.data import DataLoader
import torch
import numpy as np
import random

import torch.distributed as dist
from transformers import AutoTokenizer
from tqdm import tqdm
from typing import List, Tuple


from cosmos_rl.rollout.vllm_rollout.vllm_rollout import vLLMRollout
from cosmos_rl.utils.parallelism import ParallelDims
from cosmos_rl.utils.distributed import init_distributed
from cosmos_rl.policy.config import Config as CosmosConfig
from cosmos_rl.utils.logging import logger
from cosmos_rl.dispatcher.data import CosmosDataset
from cosmos_rl.utils import util


def main(args: argparse.Namespace):
    init_distributed()

    config = [
        None,
    ]
    local_rank = int(os.environ.get("LOCAL_RANK", 0))
    global_rank = int(os.environ.get("RANK", 0))
    world_size = int(os.environ.get("WORLD_SIZE", 1))
    device = torch.device(f"cuda:{local_rank}")
    torch.cuda.set_device(device)

    config_file = args.config
    iteration_number = args.number

    if global_rank == 0:
        try:
            logger.info(f"Attempting to load configuration from {config_file}")
            with open(config_file, "r") as f:
                config_dict = toml.load(f)

            config[0] = CosmosConfig.from_dict(config_dict)
            logger.info(f"Successfully loaded configuration from {config_file}")
        except Exception as e:
            raise RuntimeError(
                f"Failed to load or parse config file {config_file}: {e}.",
                exc_info=True,
            )
    if world_size > 1:
        dist.broadcast_object_list(config, src=0, device=torch.device("cpu"))

    config = config[0]

    task_type = config.train.train_policy.type
    assert "grpo" in task_type.lower(), "Only GRPO needs rollout."

    tokenizer = AutoTokenizer.from_pretrained(config.policy.model_name_or_path)

    seed = config.rollout.seed
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)

    parallel_dims = ParallelDims.from_config(
        parallesim_config=config.rollout.parallelism
    )
    parallel_dims.build_mesh(device_type="cuda")

    batch_size = config.rollout.batch_size

    total_step = [
        0,
    ]
    if parallel_dims.global_rank == 0:
        # only the rank 0 is responsible for loading the dataset.
        try:
            # Load GRPO dataset
            # preprocess the dataset if is cosmos. This is highly consistent with the controller.
            is_cosmos = "cosmos" in config.train.train_policy.dataset.name.lower()
            if is_cosmos:
                util.prepare_cosmos_data(config=config.train.train_policy.dataset)
            dataset = CosmosDataset(config=config)
            train_dataloader = DataLoader(dataset.train_set, batch_size=1, shuffle=True)
            logger.info(
                f"len(dataset.train_set): {len(dataset.train_set)}, batch_size: {batch_size}"
            )
            total_step[0] = (len(dataset.train_set) + batch_size - 1) // batch_size
            train_dataloader_iter = iter(train_dataloader)
        except Exception as e:
            logger.error(f"Fail to load dataset: {e}")

    if world_size > 1:
        dist.broadcast_object_list(total_step, src=0, device=torch.device("cpu"))

    total_step = total_step[0]

    logger.info(f"total_step: {total_step}, iteration_number: {iteration_number}")

    executed_step = min(iteration_number, total_step)
    if executed_step < 0:
        executed_step = total_step

    pbar = tqdm(total=executed_step, desc="[Benchmark] Step")

    rollout_engine = vLLMRollout(
        config,
        tokenizer=tokenizer,
        seed=seed,
        load_format="auto",
    )

    all_processed_tokens = 0
    time_of_all_iterations = 0

    for i in range(executed_step):
        temp = [None]
        if parallel_dims.global_rank == 0:
            pbar.update(1)
            prompt_id_and_payload_list: List[Tuple[int, str]] = []
            # get the prompt
            for _ in range(batch_size):
                try:
                    idx, prompt = next(train_dataloader_iter)
                except StopIteration:
                    logger.info("Rollout iterate over the dataset.")
                    break

                idx = idx.item() if isinstance(idx, torch.Tensor) else idx
                prompt_id_and_payload_list.append(
                    (
                        idx,
                        prompt[0]
                        if isinstance(prompt, list) or isinstance(prompt, tuple)
                        else prompt,
                    )
                )
            temp[0] = prompt_id_and_payload_list

        if world_size > 1:
            dist.broadcast_object_list(temp, src=0, device=torch.device("cpu"))

        prompt_id_and_payload_list = temp[0]

        # calculate the token number
        token_input_nums = []
        for _, prompt in prompt_id_and_payload_list:
            token_input_nums.append(len(tokenizer.encode(prompt)))

        # E2E rollout generation perf: prompt preprocessing + rollout generation
        start = time.perf_counter()
        completions_per_prompt = rollout_engine.rollout_generation(
            prompt_id_and_payload_list, stream=None
        )
        end = time.perf_counter()

        # calculate the output token number
        output_token_nums = []
        for completions in completions_per_prompt:
            token_num = 0
            for completion in completions:
                token_num += len(tokenizer.encode(completion))
            output_token_nums.append(token_num)

        total_output_tokens_num = sum(output_token_nums)
        total_tokens_num = total_output_tokens_num + sum(token_input_nums)

        all_processed_tokens += total_tokens_num
        time_of_all_iterations += end - start

        if (
            parallel_dims.tp_coord[0] == 0
            and parallel_dims.pp_coord[0] == parallel_dims.pp_coord[1] - 1
        ):
            elapsed_time = end - start
            logger.info(
                f"\nPerformance with batch size {batch_size}:\n"
                f"Total tokens/s: {total_tokens_num / elapsed_time:.2f}\n"
                f"Output tokens/s: {total_output_tokens_num / elapsed_time:.2f}"
            )
    if (
        parallel_dims.tp_coord[0] == 0
        and parallel_dims.pp_coord[0] == parallel_dims.pp_coord[1] - 1
    ):
        logger.info(
            f"All processed tokens: {all_processed_tokens} in {executed_step} iterations, time of all iterations: {time_of_all_iterations}, token/s: {all_processed_tokens / time_of_all_iterations:.2f}"
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config", type=str, help="File path of the cosmos config.", required=True
    )
    parser.add_argument(
        "--number", type=int, help="Number of iterations.", required=True
    )

    args = parser.parse_args()

    main(args)
